BeginPackage["ArAlg`","Misc`","Spectrum`"]

YuleWalker::usage = "YuleWalker[cov_] solves the Yule-Walker equations for
the given covariance sequence. Returns {sigu2, a}."

Levinson::usage = "Levinson[cov_] implements the Levinson-Durbin algorithm
for the given covariance sequence. Returns {sigu2, a}."

KtoA::usage = "KtoA[K_] implements the part of the Levinson-Durbin algorithm, 
computing the AR parameters from the given sequence of partial correlation 
coefficients."

AtoK::usage = "AtoK[a_] implements the Inverse Levinson algorithm, computing 
the sequence of partial correlation coefficients from the given AR parameters."

KtoR::usage = "KtoR[K_, sigu2_, m_] implements the Inverse Schur algorithm, 
computing the covariance sequence from the given partial correlation 
coefficients of an AR(p) process and the innovation variance. Returns {r} in 
the range 0 through m."

RtoK::usage = "RtoK[r_] implements the Schur algorithm, computing the 
partial correlation coefficients and the innovation variance from the given 
covariance sequence.  Returns {sigu2,K}."

LstSqrAr::usage = "LstSqrAr[data_, p_, opts___] computes any of the four
least-squares AR(p) estimates from data. Options and defaults: Method->FCov
is the method. FCov, the default, for Forward Covariance. BCov for Backward
Covariance, FBCov for Forward-Backward Covariance, PWin for Prewindowed.
Returns {sigu2,a}."

ParCor::usage = "ParCor[data_, p_, opts___] estimates p partial correlation 
coefficients from data by either Itakura-Saito or Burg methods. Options and 
defaults: Method->ItakuraSaito is the default method, Method->Burg is the 
option. Returns {sigu2,a,K}."

MvSpec::usage = "MvSpec[cov_, omega_] computes the minimum-variance spectral 
estimator corresponding to cov. If omega is an integer, it gives a list of 
spectrum values at omega frequency points equally spaced in the range [0, pi].  
If omega is a list of radian frequencies, it computes the spectrum values at 
those frequencies."

Method::usage = "Method is an option of some of the functions."
FCov::usage = "Method->FCov is an option for LstSqrAr."
BCov::usage = "Method->BCov is an option for LstSqrAr."
FBCov::usage = "Method->FBCov is an option for LstSqrAr."
PWin::usage = "Method->PWin is an option for LstSqrAr."
ItakuraSaito::usage = "Method->ItakuraSaito is an option for ParCor."
Burg::usage = "Method->Burg is an option for ParCor."

Begin["`Private`"]

YuleWalker[cov_List] :=
   Module[{a,n,r1,r2,sigu2},
         n = Length[cov] - 1;
         r1 = Take[cov,n];
         r2 = Take[cov,-n];
         a = Prepend[-LinearSolve[Toeplitz[r1,r1],r2], 1];
         sigu2 = cov . a;
         Return[{sigu2, a}]]

Levinson[cov_List] :=
   Module[{a={1},i,K,sigu2=cov[[1]]},
         Do[(K = (Take[cov,{2,i+1}] . Reverse[a])/sigu2;
             a = Append[a,0] - K Reverse[Append[a,0]];
             sigu2 = sigu2 (1 - K^2);), {i,1,Length[cov]-1}];
         Return[{sigu2, a}]]

KtoA[K_List] :=
   Module[{a={1},i},
         Do[a = Append[a,0] - K[[i]] Reverse[Append[a,0]],
            {i,1,Length[K]}];
         Return[a]]

AtoK[a_List] :=
   Module[{i,k,K={},p,temp,u},
         p = Length[a] - 1;
         temp = {a,Reverse[a]};
         Do[(k = -Last[temp[[1]]]; K = Prepend[K,k];
             u = (1-k^2)^(-1) {{1,k},{k,1}}; temp = u . temp;
             temp = {Drop[temp[[1]],-1],Drop[temp[[2]],1]};),
            {i,1,p}];
         Return[Simplify[K]]]

KtoR[K_List, sigu2_, m_Integer] :=
   Module[{b,f,i,j,Kc,p,r0,r,u},
         p = Length[K];
         If[p == 0, Return[Prepend[Zeros[m],sigu2]]];
         Kc = 1 - K^2;
         u = Table[{{1,K[[i]]},{-K[[i]],Kc[[i]]}},{i,1,p}];
         r0 = sigu2 / Apply[Times,Kc];
         r = {r0};
         f = b = Prepend[Zeros[p], r0];
         Do[(Do[{f[[p+1-j]],b[[p+2-j]]} = u[[p+1-j]] . {f[[p+2-j]],b[[p+1-j]]},
                {j,1,p}];
             b[[1]] = f[[1]]; r = Append[r,f[[1]]];), {i,1,m}];
         Return[Simplify[r]]]

RtoK[r_List] :=
   Module[{i,k,K={},p,sigu2,temp,u},
         p = Length[r] - 1;
         temp = {Take[r,p],Take[r,-p]};
         Do[(k = temp[[2,1]]/temp[[1,1]]; AppendTo[K,k];
             u = {{1,-k},{-k,1}}; temp = u . temp;
             If[i == p, sigu2 = temp[[1,1]],
                temp = {Drop[temp[[1]],-1],Drop[temp[[2]],1]}];),
            {i,1,p}];
         Return[{sigu2,K}]]

Options[LstSqrAr] = {Method->FCov}
LstSqrAr[data_List, p_Integer, opts___] :=
   Module[{ahat,mat,method,Nsam,sigu2,vec},
          method = Method /. {opts} /. Options[LstSqrAr];
          Nsam = Length[data];
          {mat,vec} = Which[method === FCov,
                            {Toeplitz[Take[data,{p,Nsam-1}],
                                      Reverse[Take[data,p]]],
                             Take[data,{p+1,Nsam}]},
                            method === BCov,
                            {Hankel[Take[data,{2,Nsam-p+1}],
                                    Take[data,{Nsam-p+1,Nsam}]],
                             Take[data,{1,Nsam-p}]},
                            method === FBCov,
                            {Join[Toeplitz[Take[data,{p,Nsam-1}],
                                            Reverse[Take[data,p]]],
                                   Hankel[Take[data,{2,Nsam-p+1}],
                                          Take[data,{Nsam-p+1,Nsam}]]],
                              Join[Take[data,{p+1,Nsam}],
                                   Take[data,{1,Nsam-p}]]},
                            method === PWin,
                            {Toeplitz[Prepend[Take[data,Nsam-1],0],Zeros[p]],
                             data},
                            True, (Print["Unrecognized Method"]; {{},{}})];
          If[vec == {}, Return[Null]];
          ahat = -LstSqr[mat,vec];
          vec = vec + mat . ahat;
          sigu2 = (vec . vec) / Length[vec];
          Return[{sigu2,Prepend[ahat,1]}]]

Options[ParCor] = {Method->ItakuraSaito}
ParCor[data_List, p_Integer, opts___] :=
   Module[{aa,ahat,bb,cc,eb,ef,i,k,K={},method,Nsam,sigu2 = 1.0},
         method = Method /. {opts} /. Options[ParCor];
         Nsam = Length[data];
         eb = ef = data;
         Do[(eb = Drop[eb,-1]; ef = Drop[ef,1];
             aa = eb . ef; bb = ef . ef; cc = eb . eb;
             k = If[method === Burg, 2.0 aa / (bb+cc), aa / Sqrt[bb cc]];
             AppendTo[K,k]; sigu2 *= (1.0 - k^2);
             {ef,eb} = {ef - k eb,eb - k ef};),
            {i,1,p}];
         sigu2 *= (data . data) / Length[data];
         ahat = KtoA[K];
         Return[{sigu2,ahat,K}]]

MvSpec[cov_List, omega_] :=
   Module[{a={1},i,K,mv = 1/cov[[1]],sigu2=cov[[1]]},
         Do[(K = (Take[cov,{2,i+1}] . Reverse[a])/sigu2;
             a = Append[a,0] - K Reverse[Append[a,0]];
             sigu2 = sigu2 (1 - K^2);
             mv = mv + Spectrum[{1},a,1/sigu2,omega];),
            {i,1,Length[cov]-1}];
         mv = Length[cov] (1/mv);
         Return[mv]]

End[]
EndPackage[]
Null
