BeginPackage["ArmaAlg`","Misc`","ArmaData`","ArAlg`","Rarma`",
             "Rarmad`","CRarma`","CRarmad`","Spectrum`","Opt`"]

ModYuleWalker::usage = "ModYuleWalker[rhat_, p_, q_] computes the Modified
Yule-Walker estimate, overdeterminded version, of the AR parameters of an
ARMA(p,q) process, from the given sample covariances rhat.  The number of 
terms in rhat must be at least p+q+1 (the minimum number gives the standard
MYW estimate)."

PerfMyw::usage = "PerfMyw[a_, b_, m_]  computes the normalized asymptotic 
variance of the modified (possibly over-determined) Yule-Walker estimates of 
the AR parameters of an ARMA process. m is the index of the highest-order
covariance used in the estimate. It also computes the lower bound on this 
estimate, when using the optimal weight matrix."

DurbinArma::usage = "DurbinArma[rhat_, p_, q_] estimates the MA parameters
of an ARMA(p,q) process by a variation of Durbin's MA method. rhat is the
sample covariance sequence, whose length determines the order of the long
AR process. DurbinArma[data_, l_, p_, q_] estimates the MA parameters
from the data sequence, where l is the order of the long AR process. The
returned values are {sigu2, b}."

MvrArma1::usage = "MvrArma1[rhat_, init_] computes the minimum variance 
ARMA(p,q) estimate based on the sample covariances rhat. init is the initial 
condition of {sigu2, a, b}. This version uses Sigma(theta) in the cost 
function.  The returned values are {sigu2, a, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

MvrArma2::usage = "MvrArma2[rhat_, init_] computes the minimum variance 
ARMA(p,q) estimate based on the sample covariances rhat. init is the initial 
condition of {sigu2, a, b}. This version uses Sigma hat in the cost function. 
The returned values are {sigu2, a, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

MvgArma1::usage = "MvgArma1[rhat_, init_] computes the minimum variance 
ARMA(p,q) estimate based on the sample correlation coefficients rhohat. init 
is the initial condition of {a, b}. This version uses Sigma(theta) in the cost 
function.  The returned values are {sigu2, a, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

MvgArma2::usage = "MvgArma2[rhat_, init_] computes the minimum variance 
ARMA(p,q) estimate based on the sample correlation coefficients rhohat. init 
is the initial condition of {a, b}. This version uses Sigma hat in the cost 
function.  The returned values are {sigu2, a, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

AmlArma1::usage = "AmlArma1[data_, init_] computes the time-domain approximate 
maximum likelihood ARMA(p,q) estimate from data. init is the initial condition 
of {sigu2, a, b}. The returned values are {sigu2, a, b}.  Options and default 
values: RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the 
cost function; MaxIter->200 is the maximum number of iterations in the 
nonlinear minimization. Method->Gradient is the nonlinear minimization method, 
either Davidon-Fletcher-Powell (Gradient) or conjugate directions without 
gradients (ConjDir)."

AmlArma2::usage = "AmlArma2[data_, init_] computes the frequency-domain 
approximate maximum likelihood ARMA(p,q) estimate from data. init is the 
initial condition of {sigu2, a, b}. The returned values are {sigu2, a, b}.  
Options and default values: RelAcc->10.0^(-6) is the relative accuracy of the 
minimum value of the cost function; MaxIter->200 is the maximum number of 
iterations in the nonlinear minimization. Method->Gradient is the nonlinear 
minimization method, either Davidon-Fletcher-Powell (Gradient) or conjugate 
directions without gradients (ConjDir)."

EmlArma::usage = "EmlArma[data_, init_] computes the exact maximum likelihood 
ARMA(p,q) estimate from data. init is the initial condition of {sigu2, a, b}. 
The returned values are {sigu2, a, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization."

Method::usage = "Method is an option for some of the procedures."
MaxIter::usage = "MaxIter is an option for some of the procedures."
RelAcc::usage = "RelAcc is an option for some of the procedures."
Gradient::usage = "Method->Gradient is an option for nonlinear minimization."
ConjDir::usage = "Method->ConjDir is an option for nonlinear minimization."

Begin["`Private`"]

ModYuleWalker[rhat_List, p_Integer, q_Integer] :=
   Module[{col,m,r,R,row,x},
         m = Length[rhat] - 1;
         row = If[q-p+1 >= 0, Reverse[Take[rhat,{q-p+2,q+1}]],
                  Join[Reverse[Take[rhat,{1,q+1}]],Take[rhat,{2,p-q}]]];
         col = Take[rhat,{q+1,m}];
         R = Toeplitz[col,row];
         r = Take[rhat,{q+2,m+1}];
         x = If[Length[row] == Length[col],
                Inverse[R] . r, LstSqr[R,r]];
         Return[Stabilize[Prepend[-x,1]]]]

PerfMyw[a_List, b_List, m_Integer] :=
   Module[{bnd,col,dsd,i,p,q,R,r,row},
         r = Rarma[a,b,1,m];
         p = Length[a] - 1;
         q = Length[b] - 1;
         row = If[q-p+1 >= 0, Reverse[Take[r,{q-p+2,q+1}]],
                  Join[Reverse[Take[r,{1,q+1}]],Take[r,{2,p-q}]]];
         col = Take[r,{q+1,m}];
         R = Toeplitz[col,row];
         row = Rarma[a,Convolution[b,b], 1, m-q-1];
         dsd = Toeplitz[row,row];
         bnd = Inverse[Transpose[R] . Inverse[dsd] . R];
         R = Inverse[Transpose[R] . R] . Transpose[R];
         dsd = R . dsd . Transpose[R];
         Return[{Table[Sqrt[dsd[[i]][[i]]], {i,1,p}],
                 Table[Sqrt[bnd[[i]][[i]]], {i,1,p}]}]]

DurbinArma[rhat_List, p_Integer, q_Integer] :=
   Module[{A,b,c,col,row,sigu2},
          {sigu2,c} = Levinson[rhat];
          row = If[p-q+1 >= 0, Reverse[Take[c,{p-q+2,p+1}]],
                   Join[Reverse[Take[c,p+1]],Zeros[q-p-1]]];
          col = Join[Drop[c,p],Zeros[q-1]];
          A = Toeplitz[col,row];
          col = Join[Drop[c,p+1],Zeros[q]];
          b = Stabilize[Prepend[-LstSqr[A,col],1]];
          Return[{sigu2,b}]]

DurbinArma[data_List, l_Integer, p_Integer, q_Integer] :=
   DurbinArma[CovSeq[data,l],p,q]

Options[MvrArma1] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvrArma1[rhat_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,g,iter,itmax,m,method,p,q,relacc,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvrArma1];
         itmax = MaxIter /. {opts} /. Options[MvrArma1];
         method = Method /. {opts} /. Options[MvrArma1];
         p = Length[init[[2]]]-1;
         q = Length[init[[3]]]-1;
         m = Length[rhat] - 1;
         f[y_] :=
            Module[{den,fun,num,r,s,sigma,vec},
                  s = y[[1]];
                  den = Prepend[Take[y,{2,p+1}],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[10.0^10]];
                  r = Rarma[den,num,s,m];
                  sigma = CRarma[den,num,s,m];
                  vec = r - rhat;
                  fun = vec . Inverse[sigma] . vec;
                  Return[fun]];
         g[y_] :=
            Module[{den,drdx,dsigmadx,grad,i,num,r,s,sigma,vec},
                  s = y[[1]];
                  den = Prepend[Take[y,{2,p+1}],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[Zeros[Length[y]]]];
                  r = Rarma[den,num,s,m];
                  drdx = Rarmad[den,num,s,m];
                  sigma = CRarma[den,num,s,m];
                  dsigmadx = CRarmad[den,num,s,m];
                  vec = Inverse[sigma] . (r - rhat);
                  grad = Zeros[Length[y]];
                  Do[grad[[i]] = (2 drdx[[i]] - vec . dsigmadx[[i]]) . vec,
                    {i,1,Length[y]}];
                  Return[grad]];
         yinit = Flatten[{init[[1]],Take[init[[2]],-p],Take[init[[3]],-q]}];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         ahat = Prepend[Take[ymin,{2,p+1}],1];
         bhat = Stabilize[Prepend[Take[ymin,-q],1]];
         Return[{ymin[[1]],ahat,bhat}]]

Options[MvrArma2] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvrArma2[rhat_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,g,iter,itmax,m,method,p,q,relacc,sigmai,
          yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvrArma2];
         itmax = MaxIter /. {opts} /. Options[MvrArma2];
         method = Method /. {opts} /. Options[MvrArma2];
         p = Length[init[[2]]]-1;
         q = Length[init[[3]]]-1;
         m = Length[rhat] - 1;
         sigmai = Inverse[CRarma[init[[2]],init[[3]],init[[1]],m]];
         f[y_] :=
            Module[{den,fun,num,r,vec},
                  den = Prepend[Take[y,{2,p+1}],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[10.0^10]];
                  r = Rarma[den,num,y[[1]],m];
                  vec = r - rhat;
                  fun = vec . sigmai . vec;
                  Return[fun]];
         g[y_] :=
            Module[{den,drdx,grad,num,r,vec},
                  den = Prepend[Take[y,{2,p+1}],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[Zeros[Length[y]]]];
                  r = Rarma[den,num,y[[1]],m];
                  drdx = Rarmad[den,num,y[[1]],m];
                  vec = sigmai . (r - rhat);
                  grad = 2 drdx . vec;
                  Return[grad]];
         yinit = Flatten[{init[[1]],Take[init[[2]],-p],Take[init[[3]],-q]}];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         ahat = Prepend[Take[ymin,{2,p+1}],1];
         bhat = Stabilize[Prepend[Take[ymin,-q],1]];
         Return[{ymin[[1]],ahat,bhat}]]

Options[MvgArma1] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvgArma1[rhat_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,g,iter,itmax,m,method,p,q,relacc,rhohat,
          sigu2hat,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvgArma1];
         itmax = MaxIter /. {opts} /. Options[MvgArma1];
         method = Method /. {opts} /. Options[MvgArma1];
         p = Length[init[[1]]]-1;
         q = Length[init[[2]]]-1;
         m = Length[rhat] - 1;
         rhohat = (rhat[[1]])^(-1) Rest[rhat];
         f[y_] :=
            Module[{den,fun,num,r,sigma,vec},
                  den = Prepend[Take[y,p],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[10.0^10]];
                  r = Garma[den,num,m];
                  sigma = CGarma[den,num,m];
                  vec = r - rhohat;
                  fun = vec . Inverse[sigma] . vec;
                  Return[fun]];
         g[y_] :=
            Module[{den,drdx,dsigmadx,grad,i,num,r,sigma,vec},
                  den = Prepend[Take[y,p],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[Zeros[Length[y]]]];
                  r = Garma[den,num,m];
                  drdx = Garmad[den,num,m];
                  sigma = CGarma[den,num,m];
                  dsigmadx = CGarmad[den,num,m];
                  vec = Inverse[sigma] . (r - rhohat);
                  grad = Zeros[Length[y]];
                  Do[grad[[i]] =  (2 drdx[[i]] - vec . dsigmadx[[i]]) . vec,
                    {i,1,Length[y]}];
                  Return[grad]];
         yinit = Join[Take[init[[1]],-p],Take[init[[2]],-q]];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         ahat = Prepend[Take[ymin,p],1];
         bhat = Stabilize[Prepend[Take[ymin,-q],1]];
         sigu2hat = rhat[[1]] / Rarma[ahat,bhat,1,0][[1]];
         Return[{sigu2hat,ahat,bhat}]]

Options[MvgArma2] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvgArma2[rhat_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,g,iter,itmax,m,method,p,q,relacc,rhohat,sigmai,
          sigu2hat,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvgArma2];
         itmax = MaxIter /. {opts} /. Options[MvgArma2];
         method = Method /. {opts} /. Options[MvgArma2];
         p = Length[init[[1]]]-1;
         q = Length[init[[2]]]-1;
         m = Length[rhat] - 1;
         rhohat = (rhat[[1]])^(-1) Rest[rhat];
         sigmai = Inverse[CGarma[init[[1]],init[[2]],m]];
         f[y_] :=
            Module[{den,fun,num,r,vec},
                  den = Prepend[Take[y,p],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[10.0^10]];
                  r = Garma[den,num,m];
                  vec = r - rhohat;
                  fun = vec . sigmai . vec;
                  Return[fun]];
         g[y_] :=
            Module[{den,drdx,grad,num,r,vec},
                  den = Prepend[Take[y,p],1];
                  num = Prepend[Take[y,-q],1];
                  If[!StableQ[den],Return[Zeros[Length[y]]]];
                  r = Garma[den,num,m];
                  drdx = Garmad[den,num,m];
                  vec = sigmai . (r - rhohat);
                  grad = 2 drdx . vec;
                  Return[grad]];
         yinit = Join[Take[init[[1]],-p],Take[init[[2]],-q]];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         ahat = Prepend[Take[ymin,p],1];
         bhat = Stabilize[Prepend[Take[ymin,-q],1]];
         sigu2hat = rhat[[1]] / Rarma[ahat,bhat,1,0][[1]];
         Return[{sigu2hat,ahat,bhat}]]

Options[AmlArma1] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
AmlArma1[data_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,g,iter,itmax,method,Nsam,p,q,relacc,s,
          yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[AmlArma1];
         itmax = MaxIter /. {opts} /. Options[AmlArma1];
         method = Method /. {opts} /. Options[AmlArma1];
         p = Length[init[[2]]]-1;
         q = Length[init[[3]]]-1;
         Nsam = Length[data];
         f[y_] := 
            Module[{an,bn,i,res=0.0,u,unew},
                  an = Reverse[Prepend[Take[y,p],1]];
                  bn = Take[y,-q];
                  u = Zeros[q];
                  Do[(unew = Take[data,{i-p,i}] . an  - bn . u;
                      res = res + unew^2;
                      u = Prepend[Drop[u,-1],unew];), {i,p+1,Nsam}];
                  Return[res]];
         g[y_] := 
            Module[{an,bn,duda,dudanew,dudb,dudbnew,i,j,res,u,unew},
                  an = Reverse[Prepend[Take[y,p],1]];
                  bn = Take[y,-q];
                  u = Zeros[q];
                  res = Zeros[p+q];
                  duda = Table[0,{p},{q}];
                  dudb = Table[0,{q},{q}];
                  Do[(unew = Take[data,{i-p,i}] . an  - bn . u;
                      Do[(dudanew = data[[i-j]] - bn . duda[[j]];
                          res[[j]] = res[[j]] + 2.0 dudanew unew;
                          duda[[j]] = Prepend[Drop[duda[[j]],-1],dudanew];),
                         {j,1,p}];
                      Do[(dudbnew = -u[[j]] - bn . dudb[[j]];
                          res[[p+j]] = res[[p+j]] + 2.0 dudbnew unew;
                          dudb[[j]] = Prepend[Drop[dudb[[j]],-1],dudbnew];),
                         {j,1,q}];
                      u = Prepend[Drop[u,-1],unew];), {i,p+1,Nsam}];
                  Return[res]];
         yinit = Join[Rest[init[[2]]],Rest[init[[3]]]];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         s = f[ymin]/(Nsam-p);
         ahat = Stabilize[Prepend[Take[ymin,p],1]];
         bhat = Stabilize[Prepend[Take[ymin,-q],1]];
         Return[{s,Prepend[Take[ymin,p],1],Prepend[Take[ymin,-q],1]}]]

Options[AmlArma2] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
AmlArma2[data_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,g,iter,itmax,method,Nsam,omega,p,q,relacc,
          s,yinit,ymin,z},
         relacc = RelAcc /. {opts} /. Options[AmlArma2];
         itmax = MaxIter /. {opts} /. Options[AmlArma2];
         method = Method /. {opts} /. Options[AmlArma2];
         p = Length[init[[2]]]-1;
         q = Length[init[[3]]]-1;
         Nsam = Length[data];
         period = InverseFourier[N[Exp[(I Pi/Nsam) (Range[Nsam]-1)]] data];
         period = Take[Abs[period]^2,Nsam/2];
         omega = N[(2.0 Pi/Nsam) (Range[Nsam/2]-0.5)];
         z = Exp[I omega];
         f[y_] := 
            Module[{an,bn,s0},
                  an = Prepend[Take[y,p],1];
                  bn = Prepend[Take[y,-q],1];
                  s0 = Spectrum[bn,an,1,omega];
                  Return[period . s0]];
         g[y_] := 
            Module[{an,az,bn,bz,dsdx,res,saux,s0},
                  an = Prepend[Take[y,p],1];
                  bn = Prepend[Take[y,-q],1];
                  res = Zeros[p+q];
                  s0 = Spectrum[bn,an,1,omega];
                  saux = Spectrum[bn,{1},1,omega];
                  az = Horner[an,z];
                  bz = Horner[bn,z];
                  Do[(dsdx = 2.0 Re[Exp[I (k-p) omega] az] saux;
                      res[[k]] = period . dsdx;), {k,1,p}];
                  Do[(dsdx = -2.0 Re[Exp[I (q-k) omega] / bz] s0;
                      res[[p+k]] = period . dsdx;), {k,1,q}];
                  Return[res]];
         yinit = Join[Rest[init[[2]]],Rest[init[[3]]]];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         s = 2 f[ymin]/Nsam;
         ahat = Stabilize[Prepend[Take[ymin,p],1]];
         bhat = Stabilize[Prepend[Take[ymin,-q],1]];
         Return[{s,ahat,bhat}]]

Options[EmlArma] = {RelAcc->10.0^(-5), MaxIter->200}
EmlArma[data_List, init_List, opts___] :=
   Module[{ahat,bhat,costmin,f,iter,itmax,n,Nsam,p,q,relacc,s,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[EmlArma];
         itmax = MaxIter /. {opts} /. Options[EmlArma];
         p = Length[init[[2]]]-1;
         q = Length[init[[3]]]-1;
         n = Max[p,q+1];
         Nsam = Length[data];
         f[y_] := 
            Module[{A,At,a,b,c,d,i,j,K,r,P,sigu2,u,x,res=0.0},
            a = Take[y,{2,p+1}];
            If[p > 0,If[!StableQ[Prepend[a,1]],Return[10.0^10]]];
            a = ZeroPad[a,n];
            A = Table[Which[i==1,-a[[j]],i==j+1,1,True,0],{i,1,n},{j,1,n}];
            At = Transpose[A];
            c = ZeroPad[Prepend[Take[y,-q],1],n];
            sigu2 = First[y];
            b = Prepend[Zeros[n-1],1];
            b = sigu2 Outer[Times,b,b];
            x = Zeros[n];
            r = Rarma[Prepend[a,1],{1},sigu2,n-1];
            P = Toeplitz[r,r];
            Do[(u = data[[i]] - c . x;
                K = P . c; d = c . K; K = d^(-1) A . K;
                x = A . x + u K;
                P = A . P. At - d Outer[Times, K, K] + b;
                res += Log[d] + u^2/d;), {i,1,Nsam}];
            Return[res]];
         yinit = Prepend[Join[Rest[init[[2]]],Rest[init[[3]]]],init[[1]]];
         {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax];
         s = ymin[[1]];
         ahat = Prepend[Take[ymin,{2,p+1}],1];
         bhat = If[q > 0, Stabilize[Prepend[Take[ymin,-q],1]], {1}];
         Return[{s,ahat,bhat}]]

End[]
EndPackage[]
Null
