BeginPackage["Bounds`","Misc`","Rarma`","Rarmad`","CRarma`"]

BndCov::usage = "BndCov[a_, b_, sigu2_, m_] computes the normalized bound
on the variances of the parameters of an ARMA process, estimated from the
sample covariances of orders 0 through m."

BndCovList::usage = "BndCovList[a_, b_, sigu2_, m_] is identical to BndCov
with the same parameters, except that it gives a list of the bounds for all
p + q <= i <= m."

BndCor::usage = "BndCor[a_, b_, m_] computes the normalized bound on the
variances of the parameters of an ARMA process, estimated from the sample
correlation coefficients of orders 1 through m."

BndCorList::usage = "BndCorList[a_, b_, m_] is identical to BndCor with the
same parameters, except that it gives a list of the bounds for all p + q <=
i <= m."

CRB::usage = "CRB[a_, b_, sigu2_] computes the normalized Cramer-Rao bound
on the variances of the parameters of an ARMA process."

Begin["`Private`"]

BndCov[a_List, b_List, sigu2_, m_Integer] :=
   Module[{p,q,drdx,sigma,i,temp,bnd},
         p = Length[a] - 1;
         q = Length[b] - 1;
         drdx = Rarmad[a,b,sigu2,m];
         sigma = CRarma[a,b,sigu2,m];
         temp = Inverse[drdx . Inverse[sigma] . Transpose[drdx]];
         temp = Table[Sqrt[temp[[i]][[i]]], {i,1,p+q+1}];
         bnd = {temp[[1]], Take[temp,{2,p+1}], Take[temp,{p+2,p+q+1}]};
         Return[bnd]]

BndCovList[a_List, b_List, sigu2_, m_Integer] :=
   Module[{p,q,drdx,sigma,i,j,temp,sigma1,drdx1,res,bnd},
         p = Length[a] - 1;
         q = Length[b] - 1;
         drdx = Rarmad[a,b,sigu2,m];
         sigma = CRarma[a,b,sigu2,m];
         res = {};
         Do[(sigma1 = Transpose[Take[Transpose[Take[sigma,j+1]],j+1]];
             drdx1 = Transpose[Take[Transpose[drdx],j+1]];
             temp = Inverse[drdx1 . Inverse[sigma1] . Transpose[drdx1]];
             temp = Table[Sqrt[temp[[i]][[i]]], {i,1,p+q+1}];
             bnd = {temp[[1]], Take[temp,{2,p+1}], Take[temp,{p+2,p+q+1}]};
             res = Append[res,bnd];) ,{j,p+q,m}];
         Return[res]]

BndCor[a_List, b_List, m_Integer] :=
   Module[{p,q,drdx,sigma,i,temp},
         p = Length[a] - 1;
         q = Length[b] - 1;
         drdx = Garmad[a,b,m];
         sigma = CGarma[a,b,m];
         temp = Inverse[drdx . Inverse[sigma] . Transpose[drdx]];
         temp = Table[Sqrt[temp[[i]][[i]]], {i,1,p+q}];
         Return[{Take[temp,p], Take[temp,-q]}]]

BndCorList[a_List, b_List, m_Integer] :=
   Module[{p,q,drdx,sigma,i,j,temp,sigma1,drdx1,res,bnd},
         p = Length[a] - 1;
         q = Length[b] - 1;
         drdx = Garmad[a,b,m];
         sigma = CGarma[a,b,m];
         res = {};
         Do[(sigma1 = Transpose[Take[Transpose[Take[sigma,j]],j]];
             drdx1 = Transpose[Take[Transpose[drdx],j]];
             temp = Inverse[drdx1 . Inverse[sigma1] . Transpose[drdx1]];
             temp = Table[Sqrt[temp[[i]][[i]]], {i,1,p+q}];
             bnd = {Take[temp,p], Take[temp,-q]};
             res = Append[res,bnd];) ,{j,p+q,m}];
         Return[res]]

Rarcross[a_List, b_List] :=
   Module[{p,q,ab,r},
         p = Length[a] - 1;
         q = Length[b] - 1;
         ab = Convolution[a,b];
         r = Rarma[ab,{1},1,p+q-1];
         r = Join[Reverse[Take[r,{2,p+q}]],r];
         r = Convolution[r,Convolution[Reverse[a],b]];
         Return[Take[r,{p+q+1,2*(p+q)-1}]]]

CRB[a_List, b_List, sigu2_] :=
   Module[{p,q,n,rxx,rxv,rvv,Rxx,Rxv,Rvv,mat,i,bnd},
         p = Length[a] - 1;
         q = Length[b] - 1;
         If[p > 0, (rxx = Rarma[a,{1},1,p-1]; Rxx = Toeplitz[rxx,rxx];)];
         If[q > 0, (rvv = Rarma[b,{1},1,q-1]; Rvv = Toeplitz[rvv,rvv];)];
         If[p > 0 && q > 0,
            (rxv = Rarcross[a,b];
             Rxv = Toeplitz[Reverse[Take[rxv,p]],Take[rxv,-q]];)];
         If[p > 0 && q > 0,
            mat = Transpose[Join[Transpose[Join[Rxx,-Transpose[Rxv]]],
                            Transpose[Join[-Rxv,Rvv]]]],
            If[q == 0, mat = Rxx, mat = Rvv]];
         mat = Table[Sqrt[Inverse[mat][[i]][[i]]], {i,1,p+q}];
         bnd = {Sqrt[2] sigu2, Take[mat,{1,p}], Take[mat,{p+1,p+q}]};
         Return[bnd]]

End[]
EndPackage[]
Null
