BeginPackage["CMarma`", "Partition`", "Liapunov`", "Cumulant`", "Misc`"]

CMarma::usage = "CMarma[{a_, b_}, stat_, kappa_, opts___] computes the 
covariance matrix of the sample moments of an ARMA process. {a, b} are the 
lists of ARMA parameters. stat is a list of lists of tau's (those lists can 
have different lengths). kappa is a function which computes the cumulants of 
the input process. The option NoOnes->True excludes partitions containing 
singletons, in case of a zero mean process. CMarma[b_, stat_, kappa_, opts___]
is for an MA process with paraters b."

Begin["`Private`"]

CAPower[a_?MatrixQ, c_?MatrixQ, 0, diag_] := c

CAPower[a_?MatrixQ, c_?MatrixQ, n_?PosIntQ, True] :=
   CAPower[a,c,n,True] = CAPower[a,c,n-1,True] a

CAPower[a_?MatrixQ, c_?MatrixQ, n_?PosIntQ, False] :=
   CAPower[a,c,n,False] = CAPower[a,c,n-1,False].a

Split[list_, var_] :=
   {Select[list,(FreeQ[#,var])&], Select[list,(!FreeQ[#,var])&]}

SetGen[tau:{__?NnIntQ}, sigma:{__?NnIntQ}, opts___] :=
   Module[{i,k,kk,l,ll,s,t,xparam},
           kk = Array[k,Length[tau]];
           ll = Array[l,Length[sigma]];
           s = SetGenAux[kk,ll+t,opts];
           s = Map[Function[xparam,Split[xparam,t]],s,{2}];
           Do[s = ReplaceRepeated[s,k[i]->tau[[i]]],{i,1,Length[tau]}];
           Do[s = ReplaceRepeated[s,l[i]->sigma[[i]]],{i,1,Length[sigma]}];
           s = ReplaceRepeated[s,t->0];
           Return[s]]

SetGenAux[a_List, b_List, opts___] :=
   Complement[PartitionAnySet[Join[a,b],opts],
   JoinedProduct[PartitionAnySet[a,opts],PartitionAnySet[b,opts]]]

JoinedProduct[a_List, b_List] :=
   Apply[Join,Map[Function[xparam,Map[(Join[xparam,#])&,b]],a]]

Set2SS[{tfree:{__?NnIntQ}, witht:{__?NnIntQ}},
   {a_?MatrixQ, b_?MatrixQ, c_?MatrixQ}] :=
   Module[{aa,bb,cc,diag,eye,f,i,l,m},
          diag = If[Length[a] == 1, True, False];
          l = Length[tfree]; m = Length[witht];
          eye = If[diag, {Table[1.0,{Length[a[[1]]]}]},
                   IdentityMatrix[Length[a]]];
          aa = Kron[KronPower[eye,l],KronPower[a,m]];
          bb = ImAinvB[a,b,l+m];
          f[i_] := CAPower[a,c,i,diag];
          cc = KronList[Map[f,Join[tfree,witht]]];
          Return[{aa,bb,cc}]]

UpLowSum[term_List, {a_?MatrixQ, b_?MatrixQ, c_?MatrixQ}, kappa_] :=
   Module[{aa,atemp,bb,btemp,cc,ctemp,fac,fix,i,m,s,tfree,witht},
   fac = 1; aa = bb = cc = {};
   fix[s_] := Drop[Sort[s-Min[s]],1];
   Do[(tfree = term[[i]][[1]]; witht = term[[i]][[2]];
       Which[tfree == {}, fac *= Cum[{a,b,c},fix[witht],kappa],
             witht == {}, fac *= Cum[{a,b,c},fix[tfree],kappa],
             True, (m = Min[tfree];
                    {atemp,btemp,ctemp} = Set2SS[{tfree-m,witht-m},{a,b,c}];
                    aa = Append[aa,atemp];
                    bb = Append[bb,btemp];
                    cc = Append[cc,ctemp];
                    fac *= kappa[Length[tfree]+Length[witht]];)];),
      {i,1,Length[term]}];
   Return[If[aa == {}, fac, fac*H[{aa,bb,cc}]]]]


MidSum[term_List, param_, kappa_, t0_?NnIntQ] :=
   Module[{fac,fix,i,s,res,t,tfree,witht},
   res = 0;
   fix[s_] := Drop[Sort[s-Min[s]],1];
   Do[(fac = 1;
       Do[(tfree = term[[i]][[1]]; witht = term[[i]][[2]];
           fac *= Cum[param,fix[Join[tfree,witht+t]],kappa];),
          {i,1,Length[term]}];
       res += fac;),{t,-t0,t0}];
   Return[res]]


CMone[{a_?MatrixQ, b_?MatrixQ, c_?MatrixQ}, tau:{__?NnIntQ},
          sigma:{__?NnIntQ}, kappa_, opts___] :=
   Module[{i,res,set,t0},
          res = 0;
          t0 = Max[Join[tau,sigma]];
          set = SetGen[Prepend[tau,0],Prepend[sigma,0]+t0+1,opts];
          res += Sum[UpLowSum[set[[i]],{a,b,c},kappa],{i,1,Length[set]}];
          set = SetGen[Prepend[sigma,0],Prepend[tau,0]+t0+1,opts];
          res += Sum[UpLowSum[set[[i]],{a,b,c},kappa],{i,1,Length[set]}];
          set = SetGen[Prepend[sigma,0],Prepend[tau,0],opts];
          res += Sum[MidSum[set[[i]],{a,b,c},kappa,t0],{i,1,Length[set]}];
          Return[res]]

CMone[b:{__?NorSQ}, tau:{__?NnIntQ}, sigma:{__?NnIntQ}, kappa_, opts___] :=
   Module[{i,res,set,t0},
          t0 = Length[b]+Max[Join[tau,sigma]];
          set = SetGen[Prepend[sigma,0],Prepend[tau,0],opts];
          res = Sum[MidSum[set[[i]],b,kappa,t0],{i,1,Length[set]}];
          Return[res]]

CMarma[{a:{__?NumberQ}, b:{__?NumberQ}}, stat_List, kappa_, opts___] :=
   Module[{aa,bb,cc,diag,fix,i,j,n,s,temp},
          fix[s_] := Drop[Sort[Append[s,0]-Min[Append[s,0]]],1];
          {aa,bb,cc} = Arma2SS[{a,b}];
          temp = Eigenvalues[aa];
          n = Length[temp];
          temp = Flatten[Table[Abs[temp[[i]]-temp[[j]]],{i,1,n},{j,i+1,n}]];
          diag = If[(n == 1) || (Max[temp] > 10.0^(-6)), True, False];
          {aa,bb,cc} = If[diag, SS2Diag[{aa,bb,cc}], SS2Schur[{aa,bb,cc}]];
          temp = Table[Which[j == i, 0.5*CMone[
                 {aa,bb,cc},fix[stat[[i]]],fix[stat[[i]]],kappa,opts],
                 j > i, CMone[
                 {aa,bb,cc},fix[stat[[i]]],fix[stat[[j]]],kappa,opts],
                 True, 0.0],
                 {i,1,Length[stat]},{j,1,Length[stat]}];
          Return[temp+Transpose[temp]]]

CMarma[b:{__?NorSQ}, stat_List, kappa_, opts___] :=
   Module[{fix,i,j,s,temp},
          fix[s_] := Drop[Sort[Append[s,0]-Min[Append[s,0]]],1];
          temp = Table[Which[j == i, 0.5*CMone[
                 b,fix[stat[[i]]],fix[stat[[i]]],kappa,opts],
                 j > i, CMone[
                 b,fix[stat[[i]]],fix[stat[[j]]],kappa,opts],
                 True, 0.0],
                 {i,1,Length[stat]},{j,1,Length[stat]}];
          Return[temp+Transpose[temp]]]

End[]
EndPackage[]
Null
