BeginPackage["Haf`","Misc`"]

POperator::usage = "POperator[data_, tau_, m_] computes the P operator of 
data, of order m and delay tau."

HAF::usage = "HAF[data_, tau_, delta_, m_, opts___] computes the discrete
high-order ambiguitiy function of data, of order m and delay tau. It also
estimates the coefficient a[m] and removes from the data the highest order
term, in preparation for the next HAF. delta is the sampling interval of the
data. The option Refine->True forces an accurate computation of the estimated 
a[m]; the default, Refine->False, uses simple interpolation. Returns
{Y,aest,datar}, where Y is the HAF of data, aest is the estimated a[m], and
datar is the reduced data (after removing the highest order term)."

PpEst::usage = "PpEst[data_, delta_, m_, opts___] performs complete estimation
of the phase parameters of the polynomial-phase signal data, by calling HAF a
total of m times. delta is the sampling interval of the data. The option
Refine->True forces an accurate computation of the estimated parameters; the
default, Refine->False, uses simple interpolation.Returns a, the estimated
phase polynomial."

PpData::usage = "PpData[Nsam_, a_, delta_] generates Nsam data points of a 
unit-amplitude polynomial-phase signal, whose phase polynomial is a, and
using sampling interval delta."

AMax::usage = "AMax[Nsam_, delta_, m_] gives the maximum absolute values 
of the coefficients of the polynomial of a polynomial-phase signal, such that
the estimate obtained by the HAF will be unaliased. Nsam is the number of
data samples, delta the sampling interval, and m the degree of the polynomial.
Returns the maximum absolute values of a[m]..a[1]."

Kfunc::usage = "Kfunc[m_, snr_] computes the threshold function K, defined in 
(12.7.12)."

Refine::usage = "Refine is an option for MaxDft."

Begin["`Private`"]

POperator[data_List, tau_Integer, m_Integer] := data /; m == 1
POperator[data_List, tau_Integer, m_Integer] :=
          Drop[data,tau]*Conjugate[Drop[data,-tau]] /; m == 2
POperator[data_List, tau_Integer, m_Integer] :=
         POperator[POperator[data,tau,m-1],tau,2] /; m > 2

DftPoint[data_, omega_, swtch_] :=
   Module[{a,f,g,gd,gd2,Nsam,res,W},
          W = E^(I omega);
          Nsam = Length[data];
          f[a_] := Fold[(#1 W + #2)&, 0, a];
          g = f[data];
          res = {Re[g*Conjugate[g]]};
          If[swtch > 0,
             (gd = -I*f[(Range[Nsam]-1)*data];
              res = Append[res,2.0*Re[g*Conjugate[gd]]];)];
          If[swtch > 1,
             (gd2 = -f[(Range[Nsam]-1)^2*data];
              res = Append[res,2.0*Re[g*Conjugate[gd2]+gd*Conjugate[gd]]];)];
          Return[res]]

Options[MaxDft] = {Refine->False}
MaxDft[data_, opts___] :=
   Module[{A,center,dw,left,mu,Nsam,omega,peak,refine,right,w,wn,y,yd,yd2,yn},
          refine = Refine /. {opts} /. Options[MaxDft];
          Nsam = Length[data];
          A = Abs[InverseFourier[HannWindow[Nsam]*data]];
          peak = Position[A,Max[A]][[1]][[1]]-1;
          left = (A[[Mod[peak-1,Nsam]+1]])^(0.25);
          center = (A[[peak+1]])^(0.25);
          right = (A[[Mod[peak+1,Nsam]+1]])^(0.25);
          omega = 2.0*N[Pi]*
                  (peak+0.5*(right-left)/(2.0*center-right-left))/Nsam;
          w = If[omega > N[Pi], omega-2.0*N[Pi], omega];
          If[!refine, Return[w]];
          {y,yd,yd2} = DftPoint[data,w,2];
          While[Abs[yd] > 10.0^(-10)*Abs[y],
                (mu = 2.0; yn = 0.0;
                 While[yn < y,
                       (mu = 0.5*mu;
                        dw = If[yd2 < 0.0, mu*yd/yd2, -0.001*Sign[yd]];
                        If[Abs[dw] < 10.0^(-8), Return[w]];
                        wn = w-dw;
                        {yn} = DftPoint[data,wn,0];)];
                 w = wn;
                 {y,yd,yd2} = DftPoint[data,wn,2];)];
           Return[w]]

HAF[data_List, tau_Integer, delta_Real, m_Integer, opts___] :=
   Module[{aest,Nsam,temp,y,yr},
           Nsam = Length[data];
           temp = ZeroPad[POperator[data,tau,m],Nsam];
           y = N[Sqrt[Nsam]]*InverseFourier[temp];
           aest = MaxDft[temp,opts]/(delta*(delta*tau)^(m-1)*m!);
           yr = data*E^(-I*aest*(delta*(Range[Nsam]-1))^m);
           Return[{y,aest,yr}]]

AMax[Nsam_Integer, delta_Real, m_Integer] :=
   Module[{mm},
          mm = m+1-Range[m];
          Return[N[Pi]/(delta*(delta*Floor[Nsam/mm])^(mm-1)*mm!)]]

PpData[Nsam_Integer, a_List, delta_Real] :=
   N[E^(I*Horner[a,delta*(Range[Nsam]-1)])];

PpEst[data_List, delta_Real, m_Integer, opts___] :=
   Module[{a={},aest,i,junk,temp,y},
          temp = data;
          Do[({junk,aest,temp} =
              HAF[temp,Floor[Length[data]/(m-i+1)],delta,m-i+1,opts];
              a = Append[a,aest];), {i,1,m}];
          a = Append[a,Arg[Apply[Plus,temp]]];
          Return[a]]

HannWindow[n_Integer] :=
   Module[{m,w},
          m = Floor[N[n/2]];
          w = N[0.5*(1.0+Cos[(Pi/m)*(Range[m+1]-1)])];
          w = Join[Reverse[Drop[w,1]],w];
          w = If[EvenQ[n],Drop[w,1],w];
          Return[w]]

Kfunc[m_?PosIntQ, snr_?NumberQ] :=
   Module[{i,m1q,q,res},
          res = 0.0;
          Do[(m1q = Binomial[m-1,q];
              res += N[10.0*Log[10,Sum[Binomial[m1q,i]^2*Factorial[i]*snr^(-i),
                                     {i,0,m1q}]]];),{q,0,m-1}];
          If[res <= 1000.0, res = 10.0*Log[10,10.0^(res/10.0)-1.0]];
          Return[res]]

End[]
EndPackage[]
Null
