BeginPackage["HoAlg`", "Misc`", "LinearAlgebra`MatrixManipulation`",
             "NgData`", "Cumulant`", "CMarma`", "Opt`"]

PreGianMendelMa::usage = "PreGianMendelMa[data_, q_, order_, opts___] 
prepares the lists of covariances and cumulants for the Giannakis-Mendel 
algorithm. data is the non-Gaussian data sequence. q is the order of the 
estimated MA model. order is the order of the cumulants used in the 
estimation algorithm. The option NoOnes->True should be used when the process
is known to have zero mean. Returns {cov,cum}, the lists of sample covariances
and sample cumulants."

GianMendelMa::usage = "GianMendelMa[{cov_, cum_}] implements the 
Giannakis-Mendel MA algorithm. {cov,cum} are the lists of sample covariances 
and cumulants, prepared by PreGianMendelMa. Returns the estimated MA 
polynomial b."

MvcMa::usage = "MvcMa[stat_, samplestat_, binit_, kappa_, opts___] implements 
the asymptotically minimum variance algorithm (based on (10.8.2)) for MA 
processes. stat is a list of lists of tau's, defining the moments to be used.
samplestat is a vector of statistics, compatible with stat (computed by 
SampleMom). binit is the initial value of the estimate, to be provided by
GianMendelMa. kappa is a function for computing the cumulants of the input 
process. The option MaxIter determines the maximum number of iterations; the 
default is 200. The option RelAcc determines the relative accuracy; the 
default is 10^(-6). The option NoOnes is for taking advantage of the 
zero-mean property of the given process; the default is True. The matrix 
Sigma hat is computed from binit as described in Section 10.5. The 
minimization is by conjugate directions (without gradient). Returns
{bhat, costmin, iter}, the estimated MA parameters, the minimum value of the 
cost function, and the number of iterations respectively."

InverseFilter::usage = "InverseFilter[b_, n_] computes an approximate FIR 
inverse filter of length n for the MA filter b. The computation is by 
minimization of (10.9.2). Returns the list of coefficients of the inverse 
filter."

MaxIter::usage = "MaxIter is an option for MvcMa."
RelAcc::usage = "RelAcc is an option for MvcMa."

Begin["`Private`"]

PreGianMendelMa[data:{__?NumberQ}, q_?PosIntQ, order_?PosIntQ, opts___] :=
   Module[{cov,cum,i,stat,temp},
          stat = Table[{i},{i,0,q}];
          cov = SampleCum[data,stat,opts];
          temp = Table[1,{i,1,order-1}];
          stat = Table[i*temp,{i,-q,q}];
          cum = SampleCum[data,stat,opts];
          Return[{cov,cum}]]

GianMendelMa[{cov:{__?NumberQ}, cum:{__?NumberQ}}] :=
   Module[{A,a,q},
           q = Length[cov]-1;
           A = AppendRows[Toeplitz[Join[Prepend[cum,0],Zeros[q-1]],Zeros[q]],
                     Toeplitz[Join[Join[Reverse[Drop[cov,1]],cov],Zeros[q]],
                              Prepend[Zeros[q],cov[q+1]]]];
           a = Join[cum,Zeros[q]];
           a = -LstSqr[A,a];
           Return[Prepend[Take[a,q],1]]]

Options[MvcMa] = {RelAcc->10.0^(-6), MaxIter->200, NoOnes->True}
MvcMa[stat_List, samplestat_List, binit_List, kappa_, opts___] :=
   Module[{costmin,iter,itmax,f,relacc,sigma,yinit,ymin},
          relacc = RelAcc /. {opts} /. Options[MvcMa];
          itmax = MaxIter /. {opts} /. Options[MvcMa];
          sigma = Inverse[CMarma[binit,stat,kappa,opts]];
          f[y_] :=
             Module[{c,fun,i,vec},
                    c = Table[Cum[Prepend[y,1.0],stat[[i]],kappa],
                              {i,Length[stat]}];
                    vec = c - samplestat;
                    fun = vec . sigma . vec;
                    Return[fun]];
          {ymin,costmin,iter} = Minimize2[f,Drop[binit,1],relacc,itmax];
          Return[{Prepend[ymin,1],costmin,iter}]]

InverseFilter[b:{__?NumberQ}, n_?PosIntQ] :=
   Module[{m,m1,m2},
          m = Length[b]+n-1;
          m1 = Floor[(m-1)/2];
          m2 = m-m1-1;
          Return[LstSqr[Toeplitz[Join[b,Zeros[n-1]],
                 Prepend[Zeros[n-1],First[b]]],Join[Zeros[m1],{1.0},Zeros[m2]]]]]

End[]
EndPackage[]
Null
