BeginPackage["Lattice`","Misc`"]

LatticeAr::usage = "LatticeAr[data_, p_, lambda_, rate_] implements the
Prewindowed AR unnormalized lattice algorithm.  data is the data sequence,
p is the order and lambda is the forgetting factor. rate is the rate at which
the results are collected. Returns the partial correlation coefficients K."

NorLatticeAr::usage = "NorLatticeAr[data_, p_, lambda_, rate_] implements the
Prewindowed AR normalized lattice algorithm.  data is the data sequence, p
is the order and lambda is the forgetting factor.  rate is the rate at which
the results are collected. Returns the partial correlation coefficients K."

LatticeArma::usage = "LatticeArma[data_, p_, lambda_, rate_] implements the 
Prewindowed ARMA unnormalized lattice algorithm.  data is the data sequence, p 
is the order and lambda is the forgetting factor.  rate is the rate at which
the results are collected.Returns the partial correlation coefficient
matrices K and K~."

KtoArma::usage = "KtoArma[{K_, Kt_}] computes the ARMA parameters from the 
partial correlation coefficient matrices K and Kt provided by the lattice ARMA 
algorithm.  Returns {ahat,bhat}."

Begin["`Private`"]

LatticeAr[data_List, p_Integer, lambda_, rate_Integer] :=
   Module[{d,e,et,etnew,g,gnew,K,khat,n,r,r0,rt,rtnew,t},
           r0 = data[[1]]^2;
           rt = Prepend[Zeros[p],r0];
           et = Prepend[Zeros[p],data[[1]]];
           g = gnew = Prepend[Zeros[p],1];
           etnew = rtnew = Zeros[p+1];
           d = K = Zeros[p];
           khat = {};
           Do[(e = etnew[[1]] = data[[t]];
               r0 = r = rtnew[[1]] = lambda r0 + data[[t]]^2;
               Do[(d[[n]] = lambda d[[n]] + e et[[n]] / g[[n]];
                   K[[n]] = d[[n]]/Sqrt[r rt[[n]]];
                   gnew[[n+1]] = g[[n]] - e^2 / r;
                   etnew[[n+1]] = et[[n]] - e d[[n]] / r;
                   rtnew[[n+1]] = rt[[n]] -  d[[n]]^2 / r;
                   e = e - et[[n]] d[[n]] / rt[[n]];
                   r = r - d[[n]]^2 / rt[[n]];),
                  {n,1,Min[t-1,p]}];
               g = gnew; et = etnew; rt = rtnew;
               If[Mod[t,rate] == 0, khat = Append[khat,K]];),
              {t,2,Length[data]}];
           Return[khat]]

NorLatticeAr[data_List, p_Integer, lambda_, rate_Integer] :=
   Module[{e,et,etnew,K,khat,n,r0,t},
           r0 = data[[1]]^2;
           et = Prepend[Zeros[p],data[[1]]/Sqrt[r0]];
           etnew = rtnew = Zeros[p+1];
           K = Zeros[p];
           khat = {};
           Do[(r0 = lambda r0 + data[[t]]^2;
               e = etnew[[1]] = data[[t]]/Sqrt[r0];
               Do[(K[[n]] = lambda K[[n]] Sqrt[(1-e^2)(1-et[[n]]^2)]
                          + e et[[n]];
                   etnew[[n+1]] = (et[[n]] - e K[[n]])
                                / Sqrt[(1-e^2)(1-K[[n]]^2)];
                   e = (e - et[[n]] K[[n]])
                     / Sqrt[(1-et[[n]]^2) (1-K[[n]]^2)];),
                  {n,1,Min[t-1,p]}];
               et = etnew;
               If[Mod[t,rate] == 0, khat = Append[khat,K]];),
              {t,2,Length[data]}];
           Return[khat]]

LatticeArma[data_List, p_Integer, lambda_, rate_Integer] :=
   Module[{d,e,et,etnew,g,gnew,K,Kt,khat,kthat,n,r,r0,rt,rtnew,t,u,yhat},
           r0 = {{1 + data[[1]]^2,data[[1]]^2},{data[[1]]^2,1 + data[[1]]^2}};
           rt = rtnew = d = K = Kt = Table[{{0,0},{0,0}},{p}];
           rt = Prepend[rt,r0];
           rtnew = Prepend[rtnew,{{0,0},{0,0}}];
           et = Table[{0,0},{p}];
           et = Prepend[et,{data[[1]],data[[1]]}];
           etnew = Prepend[et,{0,0}];
           g = gnew = Prepend[Zeros[p],1];
           khat = {}; kthat = {};
           Do[(u = data[[t]] - yhat;
               e = etnew[[1]] = {data[[t]],u};
               r0 = r = rtnew[[1]] = lambda r0 +
                    {{data[[t]]^2,data[[t]] u},{data[[t]] u,u^2}};
               yhat = 0;
               Do[(d[[n]] = lambda d[[n]] + g[[n]]^(-1) Outer[Times,e,et[[n]]];
                   K[[n]] = Inverse[r] . d[[n]];
                   Kt[[n]] = Inverse[rt[[n]]] . Transpose[d[[n]]];
                   gnew[[n+1]] = g[[n]] - e . Inverse[r] . e;
                   etnew[[n+1]] = et[[n]] - e . K[[n]];
                   rtnew[[n+1]] = rt[[n]] -  Transpose[d[[n]]] . K[[n]];
                   e = e - et[[n]] . Kt[[n]];
                   r = r - d[[n]] . Kt[[n]];),
                  {n,1,Min[t-1,p]}];
               yhat = Sum[etnew[[n]] . Kt[[n]],{n,1,p}][[1]];
               g = gnew; et = etnew; rt = rtnew;
               If[Mod[t,rate] == 0,
                  (khat = Append[khat,Table[Transpose[K[[n]]],{n,1,p}]];
                   kthat = Append[kthat,Table[Transpose[Kt[[n]]],{n,1,p}]];
                  )];),
              {t,2,Length[data]}];
           Return[{khat,kthat}]]

KtoArma[{K_List, Kt_List}] :=
   Module[{ahat,A,At,bhat,i,n,temp1,temp2},
         A = At = {{{1,0},{0,1}}};
         Do[(temp1 = Append[A,{{0,0},{0,0}}];
             temp2 = Prepend[At,{{0,0},{0,0}}];
             A = Table[temp1[[i]] - Kt[[n]] . temp2[[i]],
                       {i,1,Length[temp1]}];
             At = Table[temp2[[i]] - K[[n]] . temp1[[i]],
                       {i,1,Length[temp2]}];),
            {n,1,Length[K]}];
         ahat = Table[A[[n]][[1]][[1]],{n,1,Length[A]}];
         bhat = Table[A[[n]][[1]][[2]],{n,2,Length[A]}];
         bhat = Prepend[-bhat,1];
         Return[{ahat,bhat}]]

End[]
EndPackage[]
Null
