BeginPackage["Liapunov`","Misc`"]

Kron::usage = "Kron[a_, b_] returns the Kronecker product of the matrices a 
and b."

KronList::usage = "KronList[list_] returns the Kronecker product of the 
matrices in list."

KronPower::usage = "KronPower[a_, n_] returns the n-th Kronecker power of the 
matrix a."

LiapEq::usage = "LiapEq[a_, b_, c_] solves the Liapunov equation x-axb=c, 
where a is lower triangular and b is upper triangular."

ImAinvB::usage = "ImAinvB[a_, b_], where a is a list of square matrices (in 
Schur form) and b is a list of column-vector-matrices, implements a part of
eq. (d.27) (the product of the last two factors in the right-hand side).
The special case ImAinvB[a_, b_, n_] where a and b are single matrices and
n is an integer, gives the special case where a and b are to be raised to
the n-th Kronecker power. If the a's are diagonal matrices, the computation 
is much more efficient. For convenience, diagonal matrices are represented by
their diagonals only (without the zeros)."

H::usage = "H[{a_, b_, c_}], where a, b, c, are lists of matrices, implements 
eq. (d.27), with t0 = 0.  H[{a_,b_,c_}, set_] treats the case discussed in
eq. (10.4.3), where set is the set of tau's (assumed nonnegative). H[h_, set_] 
implements the special case of finite impulse response, as given by (10.4.1).
In this case h is allowed to be a list of numbers or symbols."

Arma2SS::usage = "Arma2SS[{a_, b_}] returns the state-space matrices 
corresponding to the ARMA parameters a and b."

SS2Schur::usage = "SS2Schur[{a_, b_, c_}] converts a given state-space 
realization to a complex Schur form."

SS2Diag::usage = "SS2Diag[{a_, b_, c_}] converts a given state-space 
realization to a diagonal form. It takes for granted that the matrix a is 
diagonalizable (i.e., it does not check it). The diagonal form of a is 
returned as {{lambda}}, where lambda are the diagonal entries, to save space." 

Begin["`Private`"]

(* ============ Kronecker Product Operations ============ *)

Kron[a_?MatrixQ, b_?MatrixQ] :=
   Module[{arow,f1},
          f1[arow_] := 
             Module[{brow,f2},
                     f2[brow_] := Apply[Join,Map[(# brow)&,arow]];
                     Map[f2,b]];
             Apply[Join,Map[f1,a]]]

KronList[a:{__?MatrixQ}] := a[[1]] /; Length[a] == 1
KronList[a:{__?MatrixQ}] := Kron[First[a],KronList[Rest[a]]]

KronPower[a_?MatrixQ, n_?PosIntQ] := a /; n == 1
KronPower[a_?MatrixQ, n_?PosIntQ] := Kron[a,KronPower[a,n-1]]

(* ======= Solution of Liapunov Equation in Schur Form ======= *)

BackSubstitution[a_?MatrixQ, b_?VectorQ] :=
   Module[{k,n,temp,x},
          n = Length[a];
          x = {b[[n]]/a[[n,n]]};
          Do[(temp = Take[a[[n-k]],{n-k,n}];
              x = Prepend[x,(b[[n-k]]-Rest[temp].x)/First[temp]];),
              {k,1,n-1}];
          Return[x]] /; Length[a] == Length[b]

ForSubstitution[a_?MatrixQ, b_?VectorQ] :=
   Module[{k,n,temp,x},
          n = Length[a];
          x = {b[[1]]/a[[1,1]]};
          Do[(temp = Take[a[[k]],{1,k}];
              x = Append[x,(b[[k]]-Drop[temp,-1].x)/Last[temp]];),
              {k,2,n}];
          Return[x]] /; Length[a] == Length[b]

LiapEq[a_?MatrixQ, b_?MatrixQ, c_?MatrixQ] :=
   Module[{bcol,ct,eye,k,n,x},
          n = Length[b];
          ct = Transpose[c];
          eye = IdentityMatrix[Length[a]];
          x = {ForSubstitution[eye - b[[1,1]] a, ct[[1]]]};
          Do[(bcol = Take[b[[k]],k];
              x = Append[x,ForSubstitution[eye - Last[bcol] a,
                         ct[[k]] + a.(Transpose[x].Drop[bcol,-1])]];),
             {k,2,n}];
          Return[Transpose[x]]]

(* =============== Computation of H =============== *)

ImAinvB[a_?MatrixQ, b_?MatrixQ, n_?PosIntQ] := ImAinvB[a,b,n] =
   Module[{aa,bb,cc,diag,n1,n2},
          diag = If[Length[a] == 1, True, False];
          If[diag,Return[Transpose[
             {(1.0 - Flatten[KronPower[a,n]])^(-1) Flatten[KronPower[b,n]]}]]];
          If[n == 1, Return[
             ForSubstitution[IdentityMatrix[Length[a]] - a, Flatten[b]]]];
          n1 = Floor[0.5 (n+1)]; n2 = n-n1;
          aa = KronPower[a,n1];
          bb = KronPower[a,n2];
          cc = KronPower[b,n1].KronPower[Transpose[b],n2];
          Return[Transpose[{Flatten[Transpose[LiapEq[aa,bb,cc]]]}]]]

ImAinvB[a:{__?MatrixQ}, b:{__?MatrixQ}] := ImAinvB[a,b] =
   Module[{aa,bb,cc,diag,n,n1,n2},
          diag = If[Length[a[[1]]] == 1, True, False];
          If[diag, Return[Transpose[
             {(1.0 - Flatten[KronList[a]])^(-1) Flatten[KronList[b]]}]]];
          n = Length[a];
          If[n == 1, Return[Transpose[{ForSubstitution[
             IdentityMatrix[Length[a[[1]]]] - a[[1]], Flatten[b[[1]]]]}]]];
          n1 = Floor[0.5 (n+1)]; n2 = n-n1;
          aa = KronList[Take[a,-n1]];
          bb = KronList[Take[a,n2]];
          cc = KronList[Take[b,-n1]].Transpose[KronList[Take[b,n2]]];
          Return[Transpose[{Flatten[Transpose[LiapEq[aa,bb,cc]]]}]]]

CAPower[a_?MatrixQ, c_?MatrixQ, 0, diag_] := c

CAPower[a_?MatrixQ, c_?MatrixQ, n_?PosIntQ, True] :=
   CAPower[a,c,n,True] = CAPower[a,c,n-1,True] a

CAPower[a_?MatrixQ, c_?MatrixQ, n_?PosIntQ, False] :=
   CAPower[a,c,n,False] = CAPower[a,c,n-1,False].a

H[{a:{__?MatrixQ}, b:{__?MatrixQ}, c:{__?MatrixQ}}] :=
   Module[{res},
          res = KronList[c].ImAinvB[a,b];
          Return[Chop[res[[1,1]]]]]

H[{a_?MatrixQ, b_?MatrixQ, c_?MatrixQ}, set:{__?NnIntQ}] :=
   Module[{diag,f,i,res},
          diag = If[Length[a] == 1, True, False];
          f[i_] := CAPower[a,c,i,diag];
          res = KronList[Map[f,set]].ImAinvB[a,b,Length[set]];
          Return[Chop[res[[1,1]]]]]

H[h:{__?NorSQ}, set:{__?NnIntQ}] :=
   Module[{i},Sum[Apply[Times,h[[i+set]]], {i,1,Length[h]-Max[set]}]]

(* =============== State-Space Conversions =============== *)

AppendColumns[l__?MatrixQ] := Join[l]

AppendRows[l__?MatrixQ] := Apply[Join, Transpose[{l}], {1}]

TakeRows[mat_?MatrixQ, part_] := Take[mat, part]

TakeColumns[mat_?MatrixQ, part_] := Take[#, part]& /@ mat

ReplaceSubMatrix[mat_?MatrixQ, submat_?MatrixQ, {i_?PosIntQ, j_?PosIntQ}] :=
   Module[{k,l,mm,ms,nm,ns},
          mm = Length[mat]; nm = Length[mat[[1]]];
          ms = Length[submat]; ns = Length[submat[[1]]];
          Return[Table[If[i <= k && k < i+ms && j <= l && l < j+ns,
                 submat[[k-i+1,l-j+1]], mat[[k,l]]], {k,mm}, {l,nm}]]]

Hermitian[a_?MatrixQ] := Conjugate[Transpose[a]]

Rsf2Csf[{q_?MatrixQ, t_?MatrixQ}] :=
   Module[{c,g,h,i,j,m,mu,n,qq,r,s,tt},
          n = Length[t]; m = n;
          qq = q; tt = t;
          While[m > 1,
                (s = Abs[tt[[m-1,m-1]]] + Abs[tt[[m,m]]];
                 If[s + Abs[tt[[m,m-1]]] > s,
                    (mu = Eigenvalues[Table[tt[[i,j]],{i,m-1,m},{j,m-1,m}]] - tt[[m,m]];
                     r = Sqrt[Abs[mu[[1]]]^2 + tt[[m,m-1]]^2];
                     c = mu[[1]]/r; s = tt[[m,m-1]]/r;
                     g = {{Conjugate[c],s},{-s,c}};
                     h = Conjugate[Transpose[g]];
                     tt = ReplaceSubMatrix[tt,g.TakeRows[tt,{m-1,m}],{m-1,1}];
                     tt = ReplaceSubMatrix[tt,TakeColumns[tt,{m-1,m}].h,{1,m-1}];
                     qq = ReplaceSubMatrix[qq,TakeColumns[qq,{m-1,m}].h,{1,m-1}];
                    )];
                 m = m-1;)];
          Return[{Chop[qq],Chop[tt]}]]

Arma2SS[{a:{__?NumberQ}, b:{__?NumberQ}}] :=
   Module[{aa,bb,cc,n},
          n = Max[Length[a]-1,Length[b]];
          If[n == 1,
             (aa = {{-a[[2]]}}; bb = {{1.0}}; cc = {b};),
             (aa = AppendColumns[{ZeroPad[-Rest[a],n]},
                   AppendRows[IdentityMatrix[n-1],Transpose[{Zeros[n-1]}]]];
              bb = Transpose[{Prepend[Zeros[n-1],1.0]}];
              cc = {ZeroPad[b,n]};)];
          Return[{aa,bb,cc}]]

SS2Schur[{a_?MatrixQ, b_?MatrixQ, c_?MatrixQ}] :=
   Module[{q,t},
          {q,t} = Rsf2Csf[SchurDecomposition[Transpose[N[a]]]];
          Return[{Hermitian[t], Hermitian[q].b, c.q}]]

SS2Diag[{a_?MatrixQ, b_?MatrixQ, c_?MatrixQ}] :=
   Module[{lam,q},
          {lam,q} = Eigensystem[a];
          q = Transpose[q];
          Return[{{lam}, Inverse[q].b, c.q}]]


End[]
EndPackage[]
Null
