BeginPackage["MaAlg`","Misc`","ArAlg`","ArmaData`","Rarma`",
             "Rarmad`","CRarma`","CRarmad`","Opt`"]

DurbinMa::usage = "DurbinMa[rhat_, q_] computes the Durbin MA(q) estimate from
rhat, using an intermediate AR model whose order is determined by the length of
rhat. DurbinMa[data_, q_, l_] computes the Durbin MA(q) estimate of data, using
an intermediate AR(l) model.  The returned values are {sigu2, b}."

MvrMa1::usage = "MvrMa1[rhat_, init_] computes the minimum variance MA(q) 
estimate based on the sample covariances rhat. init is the initial condition 
of {sigu2, b}. This version uses Sigma(theta) in the cost function. The 
returned values are {sigu2, b}. Options and default values: RelAcc->10.0^(-6) 
is the relative accuracy of the minimum value of the cost function; 
MaxIter->200 is the maximum number of iterations in the nonlinear minimization.
Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

MvrMa2::usage = "MvrMa2[rhat_, init_] computes the minimum variance MA(q) 
estimate based on the sample covariances rhat. init is the initial condition 
of {sigu2, b}. This version uses Sigma hat in the cost function. The returned 
values are {sigu2, b}.  Options and default values: RelAcc->10.0^(-6) is the 
relative accuracy of the minimum value of the cost function; MaxIter->200 is 
the maximum number of iterations in the nonlinear minimization. 
Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

MvgMa1::usage = "MvgMa1[rhat_, init_] computes the minimum variance MA(q) 
estimate based on the sample correlation coefficients rhohat. init is the 
initial condition of b. This version uses Sigma(theta) in the cost function.  
The returned values are {sigu2, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

MvgMa2::usage = "MvgMa2[rhat_, init_] computes the minimum variance MA(q) 
estimate based on the sample correlation coefficients rhohat. init is the 
initial condition of b. This version uses Sigma hat in the cost function.  The 
returned values are {sigu2, b}.  Options and default values: RelAcc->10.0^(-6) 
is the relative accuracy of the minimum value of the cost function; 
MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

AmlMa::usage = "AmlMa[data_, init_] computes the approximate maximum 
likelihood MA(q) estimate from data. init is the initial condition of {sigu2, 
b}. The returned values are {sigu2, b}.  Options and default values: 
RelAcc->10.0^(-6) is the relative accuracy of the minimum value of the cost 
function; MaxIter->200 is the maximum number of iterations in the nonlinear 
minimization. Method->Gradient is the nonlinear minimization method, either 
Davidon-Fletcher-Powell (Gradient) or conjugate directions without gradients 
(ConjDir)."

Method::usage = "Method is an option for some of the procedures."
MaxIter::usage = "MaxIter is an option for some of the procedures."
RelAcc::usage = "RelAcc is an option for some of the procedures."
Gradient::usage = "Method->Gradient is an option for nonlinear minimization."
ConjDir::usage = "Method->ConjDir is an option for nonlinear minimization."

Begin["`Private`"]

DurbinMa[rhat_List, q_Integer] :=
   Module[{b,sigu2},
          b = Levinson[CovSeq[Levinson[rhat][[2]],q]][[2]];
          sigu2 = rhat[[1]]/(b . b);
          Return[{sigu2,b}]]

DurbinMa[data_List, q_Integer, l_Integer] := DurbinMa[CovSeq[data,l],q]

Options[MvrMa1] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvrMa1[rhat_List, init_List, opts___] :=
   Module[{costmin,iter,itmax,f,g,m,method,q,relacc,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvrMa1];
         itmax = MaxIter /. {opts} /. Options[MvrMa1];
         method = Method /. {opts} /. Options[MvrMa1];
         q = Length[init[[2]]] - 1;
         m = Length[rhat] - 1;
         f[y_] :=
            Module[{fun,num,r,s,sigma,vec},
                  s = y[[1]];
                  num = Prepend[Take[y,-q],1];
                  r = Rarma[{1},num,s,m];
                  sigma = CRarma[{1},num,s,m];
                  vec = r - rhat;
                  fun = vec . Inverse[sigma] . vec;
                  Return[fun]];
         g[y_] :=
            Module[{drdx,dsigmadx,grad,i,num,r,s,sigma,vec},
                  s = y[[1]];
                  num = Prepend[Take[y,-q],1];
                  r = Rarma[{1},num,s,m];
                  drdx = Rarmad[{1},num,s,m];
                  sigma = CRarma[{1},num,s,m];
                  dsigmadx = CRarmad[{1},num,s,m];
                  vec = Inverse[sigma] . (r - rhat);
                  grad = Zeros[Length[y]];
                  Do[grad[[i]] = (2 drdx[[i]] - vec . dsigmadx[[i]]) . vec,
                    {i,1,Length[y]}];
                  Return[grad]];
         yinit = Flatten[{init[[1]],Take[init[[2]],-q]}];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         Return[{ymin[[1]], Prepend[Take[ymin,-q],1]}]]

Options[MvrMa2] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvrMa2[rhat_List, init_List, opts___] :=
   Module[{costmin,f,g,iter,itmax,m,method,q,relacc,sigmai,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvrMa2];
         itmax = MaxIter /. {opts} /. Options[MvrMa2];
         method = Method /. {opts} /. Options[MvrMa2];
         q = Length[init[[2]]]-1;
         m = Length[rhat] - 1;
         sigmai = Inverse[CRarma[{1},init[[2]],init[[1]],m]];
         f[y_] :=
            Module[{fun,num,r,vec},
                  num = Prepend[Take[y,-q],1];
                  r = Rarma[{1},num,y[[1]],m];
                  vec = r - rhat;
                  fun = vec . sigmai . vec;
                  Return[fun]];
         g[y_] :=
            Module[{drdx,grad,num,r,vec},
                  num = Prepend[Take[y,-q],1];
                  r = Rarma[{1},num,y[[1]],m];
                  drdx = Rarmad[{1},num,y[[1]],m];
                  vec = sigmai . (r - rhat);
                  grad = 2 drdx . vec;
                  Return[grad]];
         yinit = Flatten[{init[[1]],Take[init[[2]],-q]}];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         Return[{ymin[[1]], Prepend[Take[ymin,-q],1]}]]

Options[MvgMa1] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvgMa1[rhat_List, init_List, opts___] :=
   Module[{bhat,costmin,f,g,iter,itmax,m,method,q,relacc,rhohat,sigu2hat,
          yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvgMa1];
         itmax = MaxIter /. {opts} /. Options[MvgMa1];
         method = Method /. {opts} /. Options[MvgMa1];
         q = Length[init] - 1;
         m = Length[rhat] - 1;
         rhohat = (rhat[[1]])^(-1) Rest[rhat];
         f[y_] :=
            Module[{fun,num,r,sigma,vec},
                  num = Prepend[y,1];
                  r = Garma[{1},num,m];
                  sigma = CGarma[{1},num,m];
                  vec = r - rhohat;
                  fun = vec . Inverse[sigma] . vec;
                  Return[fun]];
         g[y_] :=
            Module[{drdx,dsigmadx,grad,i,num,r,sigma,vec},
                  num = Prepend[y,1];
                  r = Garma[{1},num,m];
                  drdx = Garmad[{1},num,m];
                  sigma = CGarma[{1},num,m];
                  dsigmadx = CGarmad[{1},num,m];
                  vec = Inverse[sigma] . (r - rhohat);
                  grad = Zeros[Length[y]];
                  Do[grad[[i]] = (2 drdx[[i]] - vec . dsigmadx[[i]]) . vec,
                    {i,1,Length[y]}];
                  Return[grad]];
         yinit = Rest[init];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         bhat = Prepend[ymin,1];
         sigu2hat = rhat[[1]] / Rarma[{1},bhat,1,0][[1]];
         Return[{sigu2hat,bhat}]]

Options[MvgMa2] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
MvgMa2[rhat_List, init_List, opts___] :=
   Module[{bhat,costmin,f,g,iter,itmax,m,method,q,relacc,rhohat,sigmai,
          sigu2hat,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[MvgMa2];
         itmax = MaxIter /. {opts} /. Options[MvgMa2];
         method = Method /. {opts} /. Options[MvgMa2];
         q = Length[init]-1;
         m = Length[rhat] - 1;
         rhohat = (rhat[[1]])^(-1) Rest[rhat];
         sigmai = Inverse[CGarma[{1},init,m]];
         f[y_] :=
            Module[{fun,num,r,vec},
                  num = Prepend[y,1];
                  r = Garma[{1},num,m];
                  vec = r - rhohat;
                  fun = vec . sigmai . vec;
                  Return[fun]];
         g[y_] :=
            Module[{drdx,grad,num,r,vec},
                  num = Prepend[y,1];
                  r = Garma[{1},num,m];
                  drdx = Garmad[{1},num,m];
                  vec = sigmai . (r - rhohat);
                  grad = 2 drdx . vec;
                  Return[grad]];
         yinit = Rest[init];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         bhat = Prepend[ymin,1];
         sigu2hat = rhat[[1]] / Rarma[{1},bhat,1,0][[1]];
         Return[{sigu2hat,bhat}]]

Options[AmlMa] = {RelAcc->10.0^(-6), MaxIter->200, Method->Gradient}
AmlMa[data_List, init_List, opts___] :=
   Module[{costmin,f,g,iter,itmax,method,Nsam,q,relacc,s,yinit,ymin},
         relacc = RelAcc /. {opts} /. Options[AmlMa];
         itmax = MaxIter /. {opts} /. Options[AmlMa];
         method = Method /. {opts} /. Options[AmlMa];
         q = Length[init[[2]]]-1;
         Nsam = Length[data];
         f[y_] := 
            Module[{i,res=0.0,u,unew},
                  u = Zeros[q];
                  Do[(unew = data[[i]] - y . u;
                      res = res + unew^2;
                      u = Prepend[Drop[u,-1],unew];), {i,1,Nsam}];
                  Return[res]];
         g[y_] := 
            Module[{dudb,dudbnew,i,j,res,u,unew},
                  u = Zeros[q];
                  res = Zeros[q];
                  dudb = Table[0,{q},{q}];
                  Do[(unew = data[[i]] - y . u;
                      Do[(dudbnew = -u[[j]] - y . dudb[[j]];
                          res[[j]] = res[[j]] + 2.0 dudbnew unew;
                          dudb[[j]] = Prepend[Drop[dudb[[j]],-1],dudbnew];),
                         {j,1,q}];
                      u = Prepend[Drop[u,-1],unew];), {i,1,Nsam}];
                  Return[res]];
         yinit = Take[init[[2]],-q];
         If[method === Gradient,
            {ymin,costmin,iter} = Minimize1[f,g,yinit,relacc,itmax],
            {ymin,costmin,iter} = Minimize2[f,yinit,relacc,itmax]];
         s = f[ymin]/Nsam;
         Return[{s,Prepend[ymin,1]}]]

End[]
EndPackage[]
Null
