BeginPackage["Misc`"]

Zeros::usage = "Zeros[n_] gives a vector of n zeros."

ZeroPad::usage = "ZeroPad[seq_, n_] zero-pads a sequence with zeros so as to 
yield a sequence of length n."

Toeplitz::usage = "Toeplitz[col_, row_] builds a Toeplitz matrix from given
first column and first row."

Hankel::usage = "Hankel[col_, row_] builds a Hankel matrix from given first
column and last row."

Convolution::usage = "Convolution[a_, b_] convolves the sequences a and b."

Horner::usage = "Horner[p_, x_] evaluates a polynomial at a point (or a list
of points) x by Horner method."

Proots::usage = "Proots[p_] finds roots of a polynomial whose coefficients are 
in the list p.  Unlike built-in MATHEMATICA routines, it returns the list of 
roots, not a list of rules for the roots."

StableQ::usage = "StableQ[p_] returns True if the polynomial p is stable,
False if it is not."

Stabilize::usage = "Stabilize[p_] stabilizes a polynomial by refelecting all 
unstable roots into the unit disk."

LstSqr::usage = "LstSqr[A_, b_] solves A x = b in the least-squares sense by
Q-R decomposition."

PosIntQ::usage = "PosIntQ[n_] is a predicate returning True if n is a 
positive integer and False otherwise."

NnIntQ::usage = "NnIntQ[n_] is a predicate returning True if n is a 
nonnegative integer and False otherwise."

NorSQ::usage = "NorSQ[x_] is a predicate returning True if x is a number or a 
symbol, and False otherwise."

Begin["`Private`"]

Zeros[n_Integer] := Table[0,{n}]

ZeroPad[seq_List, n_Integer] := Join[seq, Zeros[n-Length[seq]]]

Toeplitz[col_List, row_List] :=
   Module[{i,j},
         Table[If[i >= j, col[[i-j+1]], row[[j-i+1]]],
               {i,1,Length[col]}, {j,1,Length[row]}]]

Hankel[col_List, row_List] :=
   Module[{i,j,n},
         n = Length[col];
         Table[If[i+j <= n+1, col[[i+j-1]], row[[i+j-n]]],
               {i,1,Length[col]}, {j,1,Length[row]}]]

Convolution[a_List, b_List] :=
   Module[{i,na,nb,ta,tb},
         na = Length[a]; nb = Length[b];
         If[na <= nb,
            {ta,tb} = {a, Join[Join[Zeros[na-1],b],Zeros[na-1]]},
            {ta,tb} = {b, Join[Join[Zeros[nb-1],a],Zeros[nb-1]]}];
         Return[Table[Reverse[ta] .
                      Take[tb,{i,i+Length[ta]-1}], {i,1,na+nb-1}]]]

Horner[p_List, x_] := Fold[(#1 x + #2)&, 0, p]

Proots[p_List] :=
   Module[{xx}, Return[xx /. Solve[Horner[p,xx] == 0 ,xx]]]

StableQ[p_List] := If[Max[Abs[Proots[p]]] < 1.0,True,False]

Stabilize[p_List] := p /; Length[p] == 1
Stabilize[p_List] :=
   Module[{rall,rs,ru,xx},
         rall = Proots[p];
         ru = Select[rall, (Abs[#] > 1.0)&];
         rs = Complement[rall,ru];
         rall = Join[rs, 1.0 / ru];
         Return[Re[Reverse[CoefficientList[
                Expand[Apply[Times,xx-rall]],xx]]]]]

LstSqr[A_, b_] :=
   Module[{q,r},
         {q,r} = QRDecomposition[A];
         Return[Inverse[r] . (q . b)]]

PosIntQ[n_] := IntegerQ[n] && (n > 0)

NnIntQ[n_] := IntegerQ[n] && (n >= 0)

NorSQ[x_] := NumberQ[x] || (Head[x] === Symbol)

End[]
EndPackage[]
Null
