BeginPackage["Opt`"]

Minimize1::usage = "Minimize1[f_, g_, x0_, tol_, itmax_] performs 
multidimensional numerical minimization by Davidon-Fletcher-Powell method.
f is the cost funcion and g is its gradient. x0 is the initial value of the 
argument, and tol is the stopping tolerance. Returns {xmin,fmin,iterations}."

Minimize2::usage = "Minimize2[f_, x0_, tol_, itmax_]  performs 
multidimensional numerical minimization by Powell conjugate directions method.  
f is the cost funcion. x0 is the initial value of the argument, and tol is the 
stopping tolerance. Returns {xmin,fmin,iterations}."

Begin["`Private`"]

Bracket[f1_, ax0_, bx0_] :=
   Module[{ax,bx,cx,fa,fb,fc,fu,glimit=100.0,gold=1.618034,q,r,
          tiny=10.0^(-20),u,ulim,itit=0},
         fa = f1[ax0]; fb = f1[bx0];
         {ax,bx,fa,fb} = If[fb <= fa, {ax0,bx0,fa,fb}, {bx0,ax0,fb,fa}];         
         cx = bx + gold (bx-ax); fc = f1[cx];
         While[fb > fc,
              (r = (bx-ax) (fb-fc); q = (bx-cx) (fb-fa);
               If[q == r, Break[]];
               u = bx - ((bx-cx) q - (bx-ax) r) /
                   (2.0 Sign[q-r] Max[Abs[q-r],tiny]);
               ulim = bx + glimit (cx-bx);
               If[((bx-u) (u-cx) > 0.0),
                  (fu = f1[u];
                   If[(fu < fc),
                      ({ax,fa,bx,fb} = {bx,fb,u,fu}; Break[];),
                      If[(fu > fb),
                         ({cx,fc} = {u,fu}; Break[];)]];
                   u = cx + gold (cx-bx);
                   fu = f1[u];),
                  If[((cx-u) (u-ulim) > 0.0),
                     (fu = f1[u];
                      If[(fu < fc),
                         ({bx,cx,u,fb,fc} = {cx,u,cx + gold (cx-bx),fc,fu};
                           fu = f1[u];)];),
                     If[((u-ulim) (ulim-cx) >= 0.0),
                        (u = ulim; fu = f1[u];),
                        (u = cx + gold (cx-bx); fu = f1[u];)]]];
               {ax,bx,cx,fa,fb,fc} = {bx,cx,u,fb,fc,fu};)];
         Return[{ax,bx,cx}]]

Brent[f1_, {ax_, bx_, cx_}, tol_] :=
   Module[{a,b,cgold=0.381966,d,e,etemp,fv,fw,fx,iter,itmax=200,p,q,r,tol1,
          tol2,v,w,x,xm,zeps=10.0^(-14)},
         {a,b} = {Min[ax,cx], Max[ax,cx]}; v = w = x = bx; e = 0.0;
         fx = fv = fw = f1[x];
         For[iter = 1, iter <= itmax, iter++,
             xm = 0.5 (a+b); tol1 = tol Abs[x] + zeps; tol2 = 2.0 tol1;
             If[Abs[x-xm] <= tol2 - 0.5 (b-a), Break[]];
             If[Abs[e] > tol1,
                (r = (x-w) (fx-fv); q = (x-v) (fx-fw); p = (x-v) q - (x-w) r;
                 q = 2.0 (q-r); p = -p Sign[q]; q = Abs[q]; etemp = e; e = d;
                 If[(Abs[p] >= Abs[0.5 q etemp]) || (p <= q (a-x)) ||
                    (p >= q (b-x)),
                    (e = If[x >= xm, a-x, b-x]; d = cgold e;),
                    (d = p/q; u = x+d;
                     If[(u-a < tol2) || ((b-u) < tol2),
                         d = tol1 Sign[xm-x]];)];),
                (e = If[x >= xm, a-x, b-x]; d = cgold e;)];
             u = If[Abs[d] >= tol1, x+d, x + tol1 Sign[d]];
             fu = f1[u];
             If[fu <= fx,
                (If[u >= x, a = x, b = x];
                 {v,fv,w,fw,x,fx} = {w,fw,x,fx,u,fu};),
                (If[u < x, a = u, b = u];
                 If[(fu <= fw) || (w == x), {v,fv,w,fw} = {w,fw,u,fu},
                    If[(fu <= fv) || (v == x) || (v == w),
                       {v,fv} = {u,fu}]];)];];
        Return[{x,fx}]]

LineMin[f_, xinit_, d_] := 
   Module[{aa,bb,cc,first,fone},
         fone = (f[xinit + # d])&;
         first = Min[1.0,0.002/Sqrt[d . d]];
         {aa,bb,cc} = Bracket[fone,0.0,first];
         Brent[fone,{aa,bb,cc},10.0^(-3)]]

Minimize1[f_, g_, x0_, tol_, itmax_] :=
   Module[{crit1,crit2,dg,dir,eps=10.0^(-8),fac1,fac2,fn,fp,gn,gp,hdg,
          Hess,i,iter,j,n,tn,xn},   
         n = Length[x0]; fp = f[x0]; fn = fp; gn = g[x0]; dir = -gn; xn = x0;
         Hess = Table[If[i == j, 1.0, 0.0], {i,1,n},{j,1,n}];
         For[iter = 1, iter <= itmax, iter++,
             (crit1 = 0.01 Sqrt[gn . gn] - tol (Abs[fp] + eps);
              If[crit1 <= 0.0, Break[]];
              {tn,fn} = LineMin[f,xn,dir]; xn = xn + tn dir;
              crit2 = 2.0 Abs[fn-fp] - tol (Abs[fn] + Abs[fp] + eps);
              If[crit2 <= 0.0, Break[]];
              fp = fn; gp = gn; gn = g[xn]; dg = gn-gp; hdg = Hess . dg; 
              fac1 = dir . dg; fac2 = dg . hdg;
              If[fac1 == 0.0 || fac2 == 0.0, Break[]];
              fac1 = 1.0 / fac1; fac2 = 1.0 / fac2;
              Hess = Hess + fac1 Outer[Times,dir,dir] -
                            fac2 Outer[Times,hdg,hdg]; 
              dir = -Hess . gn;)];
         Return[{xn, fn, iter}]]

Minimize2[f_, x0_, tol_, itmax_] :=
   Module[{crit1,crit2,del,dirset,dirt,eps=10.0^(-8),fn,fp,ftt,i,ibig,iter,
          j,n,tn,x,xt,xtt},
         n = Length[x0]; fn = f[x0]; x = x0; xt = x0;
         dirset = Table[If[i == j, 1.0, 0.0], {i,1,n},{j,1,n}];
         For[iter = 1, iter <= itmax, iter++,
             (fp = fn; ibig = 0; del = 0.0;
              Do[(dirt = dirset[[i]]; {tn,fn} = LineMin[f,x,dirt];
                  x = x + tn dirt;
                  If[Abs[fp-fn] > del, (del = Abs[fp-fn]; ibig = i;)];),
                 {i,1,n}];
              crit1 = 2.0 Abs[fp-fn] - tol (Abs[fp] + Abs[fn] + eps);
              If[crit1 <= 0.0, Break[]];
              xtt = 2.0 x - xt; dirt = x - xt; xt = x;
              ftt = f[xtt]; If[ftt >= fp, Continue[]];
              crit2 = 2.0 (fp - 2.0 fn + ftt) (fp-fn-del)^2 - del (fp-ftt)^2;
              If[crit2 >= 0.0, Continue];
              {tn,fn} = LineMin[f,x,dirt];
              x = x + tn dirt; dirset[[ibig]] = dirt;)];
         Return[{x,fn,iter}]]

End[]
EndPackage[]
Null
