BeginPackage["Period`","Misc`"]

Periodogram::usage = "Periodogram[data_] computes the periodogram of the
sequence data."

AveragedPeriodogram::usage = "AveragedPeriodogram[data_, window_, overlap_]
computes the averaged periodogram of the sequence data. window is the desired 
window (e.g. a Hann window for the standard Welch method). The length of the
window determines the length of the periodogram. overlap is the overlap
parameter L (an integer), which should be less than the length of the
periodogram. A rectangular window with zero overlap gives the Bartlett method."

SmoothedPeriodogram::usage = "SmoothedPeriodogram[data_, window_] computes the
smoothed periodogram of the sequence data. window is the set of weights, 
assumed to be symmetric." 

WindowedPeriodogram::usage = "WindowedPeriodogram[data_, window_, npoint_]
computes the windowed periodogram of the sequence data. window is the desired 
window. npoint is the desired number of frequency points."

RectangularWindow::usage = "RectangularWindow[n_] returns a rectangular
window of length n."

TriangularWindow::usage = "TriangularWindow[n_] returns a triangular window of
length n."

HannWindow::usage = "HannWindow[n_] returns a Hann (cosine) window of length n."

HammingWindow::usage = "HannWindow[n_] returns a Hamming (raised cosine) window
of length n."

BlackmanWindow::usage = "BlackmanWindow[n_] returns a Blackman window of
length n."

KaiserWindow::usage = "KaiserWindow[n_, beta_] returns a Kaiser window of
length n and parameter beta."

ConvWindow::usage = "ConvWindow[window_] convolves a window with itself."

Begin["`Private`"]

Periodogram[data_List] := (Abs[Fourier[data]])^2

AveragedPeriodogram[data_List, window_List, overlap_Integer] :=
   Module[{i,n,n1,n2,p,w,y},
           n = Length[data];
           n1 = Length[window];
           w = N[Sqrt[n1/(window.window)]*window];
           n2 = Floor[N[(n-overlap)/(n1-overlap)]];
           p = Zeros[n1];
           Do[(y = Take[data,{i*(n1-overlap)+1,i*(n1-overlap)+n1}];
               p = p+Periodogram[w*y];), {i,0,n2-1}];
           Return[(1.0/n2)*p]]

SmoothedPeriodogram[data_List, window_List] :=
   Module[{m,m1,n,w},
          n = Length[data];
          m = Length[window];
          m1 = Floor[N[m/2]] + 1;
          w = Join[Join[Take[window,{m1,m}],Zeros[n-2*(m-m1)-1]],
               Reverse[Take[window,{m1+1,m}]]];
          w = (1.0/Apply[Plus,w])*w;
          Return[N[Sqrt[n]]*
                 Re[InverseFourier[Fourier[Periodogram[data]]*Fourier[w]]]]]

WindowedPeriodogram[data_List, window_List, npoint_Integer] :=
   Module[{m,m1,r},
          m = Length[window];
          m1 = Floor[N[m/2]] + 1;
          r = Join[CovSeq[data,m-m1]*Take[window,{m1,m}],
                   Zeros[npoint-(m-m1+1)]];
          r = Prepend[Drop[r,1],0.5*r[[1]]];
          Return[2.0*N[Sqrt[npoint]]*Re[Fourier[r]]]]

RectangularWindow[n_Integer] := Table[1.0,{n}]

TriangularWindow[n_Integer] :=
   Module[{m,w},
          m = Floor[N[n/2]];
          w = (1.0/m)*Reverse[(Range[m+1]-1)];
          w = Join[Reverse[Drop[w,1]],w];
          w = If[EvenQ[n],Drop[w,1],w];
          Return[w]]

HannWindow[n_Integer] :=
   Module[{m,w},
          m = Floor[N[n/2]];
          w = N[0.5*(1.0+Cos[(Pi/m)*(Range[m+1]-1)])];
          w = Join[Reverse[Drop[w,1]],w];
          w = If[EvenQ[n],Drop[w,1],w];
          Return[w]]

HammingWindow[n_Integer] :=
   Module[{m,w},
          m = Floor[N[n/2]];
          w = N[0.54+0.46*Cos[(Pi/m)*(Range[m+1]-1)]];
          w = Join[Reverse[Drop[w,1]],w];
          w = If[EvenQ[n],Drop[w,1],w];
          Return[w]]

BlackmanWindow[n_Integer] :=
   Module[{m,w},
          m = Floor[N[n/2]];
          w = N[0.42+0.5*Cos[(Pi/m)*(Range[m+1]-1)]
                    +0.08*Cos[(2*Pi/m)*(Range[m+1]-1)]];
          w = Join[Reverse[Drop[w,1]],w];
          w = If[EvenQ[n],Drop[w,1],w];
          Return[w]]

KaiserWindow[n_Integer, beta_Real] :=
   Module[{m,w},
          m = Floor[N[n/2]];
          w = BesselI[0,beta*N[Sqrt[1 - ((1/m)*(Range[m+1]-1))^2]]];
          w = Join[Reverse[Drop[w,1]],w];
          w = (1.0/BesselI[0,beta])*w;
          w = If[EvenQ[n],Drop[w,1],w];
          Return[w]]

ConvWindow[window_] := ((1.0)/(window.window))*Convolution[window,window]

CovSeq[data_List, nr_Integer] :=
   Module[{i,ndat,r},
         ndat = Length[data];
         r = Zeros[nr+1];
         Do[r[[i+1]] = Take[data,ndat-i] . Take[data,-(ndat-i)],
            {i,0,nr}];
         Return[r/ndat]]

End[]
EndPackage[]
Null
