BeginPackage["Rls`","Misc`","LinearAlgebra`MatrixManipulation`"]

Rls::usage = "Rls[data_, p_, q_, lambda_, rate_, opts___] implements the
following three algorithms: the recursive least-squares for AR estimation,
the extended least-squares for ARMA estimation, and the recursive maximum
likelihood for ARMA estimation. data is the data sequence, p and q are the AR
and MA orders respectively.  The special case q = 0 gives the AR/RLS algorithm.
lambda is the forgetting factor. rate is the rate at which the results are
collected. Options and defaults: Method->RML chooses the algorithms for the
ARMA case, RML (default) or ELS. ForceStab->False chooses whether to force a
stable b(z) at each iteration. Returns {ahat,bhat}."

RlsSqrt::usage = "RlsSqrt[data_, p_, q_, lambda_, rate_, opts___] is
identical to Rls, except that it implements the P matrix update in
square-root form." 
 
StGrad::usage = "StGrad[data_, p_, q_, mu_, rate_, opts___] implements the
stochastic gradient algorithm for AR or ARMA estimation. data is the data
sequence, p and q are the AR and MA orders respectively.  The special case
q = 0 gives the AR algorithm. mu determines the step size, namely: mu itself
in the LMS version, or mu/t in the Robbins-Monro version. rate is the rate at
which the results are collected. Options and defaults: Method->LMS chooses
the algorithm, LMS (default) or Robbins-Monro (RM). ForceStab->False chooses
whether to force a stable b(z) at each iteration. Returns {ahat,bhat}."
 
Method::usage = "Method is an option for Rls and StGrad."
ForceStab::usage = "ForceStab is an option for Rls and StGrad."
RML::usage = "Method->RML is an option for Rls."
LMS::usage = "Method->LMS is an option for StGrad."

Begin["`Private`"]

Options[Rls] = {Method->RML, ForceStab->False}
Rls[data_List, p_Integer, q_Integer, lambda_, rate_Integer, opts___] :=
   Module[{ae,ahat,be,bhat,i,j,K,method,Nsam,P,stab,t,theta,u,new,v,
          vnew,w,wnew,y,ynew},
         method = Method /. {opts} /. Options[Rls];
         stab = ForceStab /. {opts} /. Options[Rls];
         Nsam = Length[data];
         P = Table[If[i==j, 10.0^(6), 0.0], {i,1,p+q}, {j,1,p+q}];
         theta = Zeros[p+q];
         u = w = Zeros[q];
         v = Zeros[Max[p,q]];
         y = Zeros[p];
         ahat = {}; bhat = {};
         Do[(eta = Join[y,u]; ynew = data[[t]]; unew = ynew - eta . theta;
             If[q > 0, u = Prepend[Drop[u,-1],unew]];
             If[p > 0, y = Prepend[Drop[y,-1],ynew]];
             If[(q > 0) && (method === RML),
                (eta = If[p==0, w, Join[Take[v,p],w]];
                 vnew = ynew - v . Take[theta,-q];
                 wnew = unew - w . Take[theta,-q];
                 v = Prepend[Drop[v,-1],vnew];
                 w = Prepend[Drop[w,-1],wnew];)];
             K = P . eta; K = (lambda + eta . K)^(-1) K;
             P = lambda^(-1) (P - Outer[Times,K,(eta . P)]);
             theta = theta + unew K;
             If[stab,(be = Stabilize[Prepend[Take[theta,-q],1]];
                      theta = Join[Take[theta,p],Take[be,-q]];)];
             If[Mod[t,rate] == 0,
                (ae = Stabilize[Prepend[-Take[theta,p],1]];
                 be = Stabilize[Prepend[Take[theta,-q],1]];
                 ahat = Append[ahat,ae]; bhat = Append[bhat,be];)];),
            {t,1,Nsam}];
         Return[{ahat,bhat}]]

Options[RlsSqrt] = {Method->RML, ForceStab->False}
RlsSqrt[data_List, p_Integer, q_Integer, lambda_, rate_Integer, opts___] :=
   Module[{ae,ahat,be,bhat,i,j,K,l2,method,Nsam,P,pq,R,stab,t,theta,u,new,v,
          vnew,w,wnew,y,ynew},
         method = Method /. {opts} /. Options[RlsSqrt];
         stab = ForceStab /. {opts} /. Options[RlsSqrt];
         Nsam = Length[data];
         pq = p+q;
         P = Table[If[i==j, 10.0^(3), 0.0], {i,1,pq}, {j,1,pq}];
         l2 = Sqrt[lambda];
         theta = Zeros[pq];
         u = w = Zeros[q];
         v = Zeros[Max[p,q]];
         y = Zeros[p];
         ahat = {}; bhat = {};
         Do[(eta = Join[y,u]; ynew = data[[t]]; unew = ynew - eta . theta;
             If[q > 0, u = Prepend[Drop[u,-1],unew]];
             If[p > 0, y = Prepend[Drop[y,-1],ynew]];
             If[(q > 0) && (method === RML),
                (eta = If[p==0, w, Join[Take[v,p],w]];
                 vnew = ynew - v . Take[theta,-q];
                 wnew = unew - w . Take[theta,-q];
                 v = Prepend[Drop[v,-1],vnew];
                 w = Prepend[Drop[w,-1],wnew];)];
             R = AppendColumns[AppendRows[{{l2}},{Zeros[pq]}],
                               AppendRows[P.Transpose[{eta}],P/l2]];
             R = QRDecomposition[R][[2]];
             P = TakeMatrix[R,{2,2},{pq+1,pq+1}];
             K = l2*Flatten[TakeMatrix[R,{1,2},{1,pq+1}]]/
                    Flatten[TakeMatrix[R,{1,1},{1,1}]][[1]];
             theta = theta + unew K;
             If[stab,(be = Stabilize[Prepend[Take[theta,-q],1]];
                      theta = Join[Take[theta,p],Take[be,-q]];)];
             If[Mod[t,rate] == 0,
                (ae = Stabilize[Prepend[-Take[theta,p],1]];
                 be = Stabilize[Prepend[Take[theta,-q],1]];
                 ahat = Append[ahat,ae]; bhat = Append[bhat,be];)];),
            {t,1,Nsam}];
         Return[{ahat,bhat}]]

Options[StGrad] = {Method->LMS, ForceStab->False}
StGrad[data_List, p_Integer, q_Integer, mu_, rate_Integer, opts___] :=
   Module[{ae,ahat,be,bhat,i,j,method,mut,Nsam,stab,t,theta,u,new,v,vnew,
          w,wnew,y,ynew},
         method = Method /. {opts} /. Options[StGrad];
         stab = ForceStab /. {opts} /. Options[StGrad];
         Nsam = Length[data];
         theta = Zeros[p+q];
         u = w = Zeros[q];
         v = Zeros[Max[p,q]];
         y = Zeros[p];
         ahat = {}; bhat = {};
         Do[(eta = Join[y,u]; ynew = data[[t]]; unew = ynew - eta . theta;
             If[q > 0, u = Prepend[Drop[u,-1],unew]];
             If[p > 0, y = Prepend[Drop[y,-1],ynew]];
             If[(q > 0),
                (eta = If[p==0, w, Join[Take[v,p],w]];
                 vnew = ynew - v . Take[theta,-q];
                 wnew = unew - w . Take[theta,-q];
                 v = Prepend[Drop[v,-1],vnew];
                 w = Prepend[Drop[w,-1],wnew];)];
             theta = theta + If[method === LMS,1,t^(-1)] 2 mu unew eta;
             If[stab,(bhat = Stabilize[Prepend[Take[theta,-q],1]];
                      theta = Join[Take[theta,p],Take[bhat,-q]];)];
             If[Mod[t,rate] == 0,
                (ae = Stabilize[Prepend[-Take[theta,p],1]];
                 be = Stabilize[Prepend[Take[theta,-q],1]];
                 ahat = Append[ahat,ae]; bhat = Append[bhat,be];)];),
            {t,1,Nsam}];
         Return[{ahat,bhat}]]

End
EndPackage[]
Null
