BeginPackage["SinAlg`","Misc`","Opt`","SinData`", "Period`", 
             "Algebra`SymbolicSum`"]

SinCrb::usage = "SinCrb[a_, phi_, omega_, n_] Computes the Cramer-Rao bound 
for multiple sinusoids in noise. a is the vector of amplitudes, phi the vector 
of phases, omega the vector of frequencies and n the number of data points
(or a list of numbers of data points). The noise variance is assumed 1 for
convenience, or take a as the ratio of the amplitude to the noise standard
deviation. Returns crb, a list of the corresponding square roots of the
diagonal elements of the inverse Fisher information matrix. If n is a list, a 
list of lists is returned."

LocalPeaks::usage = "LocalPeaks[p_, n_] finds the n largest local peaks in
the periodogram p. Returns the list of frequencies and the corresponding
lists of periodogram ordinates. LocalPeaks[data_, zpfactor_, window_, n_]
is similar, but accepts the following inputs: data, the data sequence;
zpfactor, an integer factor for zero-padding the data; window, a function
name for optional windowing of the data before computing the periodogram
(must be one of the window functions in Period.m)."

SinMl::usage = "SinMl[data_, init_, opts___] computes the maximum likelihood 
estimates for multiple sinusoids in white noise. data is the data sequence, 
init is the initial condition for the frequencies. The nonlinear minimization 
is by conjugate directions (without gradient). Options and defaults values: 
Method->Exact chooses the type of maximum likeliood; Exact is the default, 
Approx the alternative. Returns {ahat, phihat, omegahat, sigmahat}."

TruncatedSvd::usage = "TruncatedSvd[data_, l_, m_, opts___] estimates the 
parameters of multiple sinusoids in white noise by the truncated SVD 
(Tufts-Kumaresan) method. data is the data sequence, l the order of the 
linear prediction polynomial, m the number of sinusoids. Options and defaults:
Method->FCov is the method of constructing the data matrix. FCov, the default,
for Forward Covariance. BCov for Backward Covariance, FBCov for Forward-
Backward Covariance, PWin for Prewindowed. Returns {ahat, phihat, omegahat,
sigmahat}."

ModYuleWalker::usage = "ModYuleWalker[data_, n_, l_, m_] estimates the
parameters of multiple sinusoids in white noise by the Modified Yule-Walker
method. data is the data sequence, n the number of sample covariances (must
be 2l at least), l the degree of the polynomial, m the number of sinusoids.
Returns {ahat, phihat. omegahat, sigmahat}."

Music::usage = "Music[data_, l_, m_] estimates the parameters of multiple
sinusoids in white noise by the MUSIC method. data is the data sequence,
l+1 the dimension of the sample covariance matrix, m the number of sinusoids.
The special case l = 2m gives the Pisarenko estimate. Returns {ahat, omegahat,
sigmahat}."

RelAcc::usage = "RelAcc->10.0^(-6) is an option for SinMl."
MaxIter::usage = "MaxIter->200 is an option for SinMl."
Method::usage = "Method"
Exact::usage = "Exact is an option for SinMl."
Approx::usage = "Approx is an option for SinMl."
FCov::usage = "Method->FCov is an option for TruncatedSvd."
BCov::usage = "Method->BCov is an option for TruncatedSvd."
FBCov::usage = "Method->FBCov is an option for TruncatedSvd."

Begin["`Private`"]

TheSum["aa", wk_, wl_, pk_, pl_, n_] := TheSum["aa",wk,wl,pk,pl,n] =
   Module[{t},
          SymbolicSum[Cos[wk*t-pk]*Cos[wl*t-pl], {t,-(n-1)/2,(n-1)/2}]]

TheSum["aa", wk_, pk_, n_] := TheSum["aa",wk,pk,n] =
   Module[{t},
          SymbolicSum[(Cos[wk*t-pk])^2, {t,-(n-1)/2,(n-1)/2}]]

TheSum["ap", wk_, wl_, pk_, pl_, n_] := TheSum["ap",wk,wl,pk,pl,n] =
   Module[{t},
          SymbolicSum[Cos[wk*t-pk]*Sin[wl*t-pl], {t,-(n-1)/2,(n-1)/2}]]

TheSum["ap", wk_, pk_, n_] := TheSum["ap",wk,pk,n] =
   Module[{t},
          SymbolicSum[Cos[wk*t-pk]*Sin[wk*t-pk], {t,-(n-1)/2,(n-1)/2}]]

TheSum["aw", wk_, wl_, pk_, pl_, n_] := TheSum["aw",wk,wl,pk,pl,n] =
   Module[{t},
          SymbolicSum[t*Cos[wk*t-pk]*Sin[wl*t-pl],
                      {t,-(n-1)/2,(n-1)/2}]]

TheSum["aw", wk_, pk_, n_] := TheSum["aw",wk,pk,n] =
   Module[{t},
          SymbolicSum[t*Cos[wk*t-pk]*Sin[wk*t-pk],
                      {t,-(n-1)/2,(n-1)/2}]]

TheSum["pp", wk_, wl_, pk_, pl_, n_] := TheSum["pp",wk,wl,pk,pl,n] =
   Module[{t},
          SymbolicSum[Sin[wk*t-pk]*Sin[wl*t-pl], {t,-(n-1)/2,(n-1)/2}]]

TheSum["pp", wk_, pk_, n_] := TheSum["pp",wk,pk,n] =
   Module[{t},
          SymbolicSum[(Sin[wk*t-pk])^2, {t,-(n-1)/2,(n-1)/2}]]

TheSum["pw", wk_, wl_, pk_, pl_, n_] := TheSum["pw",wk,wl,pk,pl,n] =
   Module[{t},
          SymbolicSum[t*Sin[wk*t-pk]*Sin[wl*t-pl],
                      {t,-(n-1)/2,(n-1)/2}]]

TheSum["pw", wk_, pk_, n_] := TheSum["pw",wk,pk,n] =
   Module[{t},
          SymbolicSum[t*(Sin[wk*t-pk])^2,
                      {t,-(n-1)/2,(n-1)/2}]]

TheSum["ww", wk_, wl_, pk_, pl_, n_] := TheSum["ww",wk,wl,pk,pl,n] =
   Module[{t},
          SymbolicSum[t^2*Sin[wk*t-pk]*Sin[wl*t-pl],
                      {t,-(n-1)/2,(n-1)/2}]]

TheSum["ww", wk_, pk_, n_] := TheSum["ww",wk,pk,n] =
   Module[{t},
          SymbolicSum[t^2*(Sin[wk*t-pk])^2,
                      {t,-(n-1)/2,(n-1)/2}]]

AppendColumns[l__?MatrixQ] := Join[l]

AppendRows[l__?MatrixQ] := Apply[Join, Transpose[{l}], {1}] 

SinCrb[a_List, phi_List, omega_List, Nsam_] :=
   Module[{fisher,flag,Jaa,Jap,Jaw,Jpp,Jpw,Jww,k,l,m,n,nlist,ntot,pk,pl,res,
           saak,saakl,sapk,sapkl,sawk,sawkl,sppk,sppkl,spwk,spwkl,swwk,swwkl,
           t,wk,wl},
          m = Length[a];
          res = {};
          If[Head[Nsam] === List, (nlist = Nsam; ntot = Length[Nsam]),
                                  (nlist = {Nsam}; ntot = 1)];
          saak = TheSum["aa",wk,pk,n];
          saakl = TheSum["aa",wk,wl,pk,pl,n];
          sapk = TheSum["ap",wk,pk,n];
          sapkl = TheSum["ap",wk,wl,pk,pl,n];
          sawk = TheSum["aw",wk,pk,n];
          sawkl = TheSum["aw",wk,wl,pk,pl,n];
          sppk = TheSum["pp",wk,pk,n];
          sppkl = TheSum["pp",wk,wl,pk,pl,n];
          spwk = TheSum["pw",wk,pk,n];
          spwkl = TheSum["pw",wk,wl,pk,pl,n];
          swwk = TheSum["ww",wk,pk,n];
          swwkl = TheSum["ww",wk,wl,pk,pl,n];
          Do[(
          Jaa = Table[If[k==l,
                      (saak //. {wk->omega[[k]],pk->phi[[k]],n->nlist[[i]]}),
                      (saakl //. {wk->omega[[k]],wl->omega[[l]],
                       pk->phi[[k]],pl->phi[[l]],n->nlist[[i]]})],
                      {k,1,m},{l,1,m}];
          Jap = Table[If[k==l,
                      a[[k]]*(sapk //. {wk->omega[[k]],pk->phi[[k]],n->nlist[[i]]}),
                      a[[l]]*(sapkl //. {wk->omega[[k]],wl->omega[[l]],
                              pk->phi[[k]],pl->phi[[l]],n->nlist[[i]]})],
                      {k,1,m},{l,1,m}];
          Jaw = Table[If[k==l,
                      -a[[k]]*(sawk //. {wk->omega[[k]],pk->phi[[k]],n->nlist[[i]]}),
                      -a[[l]]*(sawkl //. {wk->omega[[k]],wl->omega[[l]],
                               pk->phi[[k]],pl->phi[[l]],n->nlist[[i]]})],
                      {k,1,m},{l,1,m}];
          Jpp = Table[If[k==l,
                      a[[k]]^2*(sppk //. {wk->omega[[k]],pk->phi[[k]],n->nlist[[i]]}),
                      a[[k]]*a[[l]]*(sppkl //. {wk->omega[[k]],wl->omega[[l]],
                                     pk->phi[[k]],pl->phi[[l]],n->nlist[[i]]})],
                      {k,1,m},{l,1,m}];
          Jpw = Table[If[k==l,
                      -a[[k]]^2*(spwk //. {wk->omega[[k]],pk->phi[[k]],n->nlist[[i]]}),
                      -a[[k]]*a[[l]]*(spwkl //. {wk->omega[[k]],wl->omega[[l]],
                                     pk->phi[[k]],pl->phi[[l]],n->nlist[[i]]})],
                      {k,1,m},{l,1,m}];
          Jww = Table[If[k==l,
                      a[[k]]^2*(swwk //. {wk->omega[[k]],pk->phi[[k]],n->nlist[[i]]}),
                      a[[k]]*a[[l]]*(swwkl //. {wk->omega[[k]],wl->omega[[l]],
                                     pk->phi[[k]],pl->phi[[l]],n->nlist[[i]]})],
                      {k,1,m},{l,1,m}];
          fisher = AppendColumns[AppendRows[Jaa,Jap,Jaw],
                   AppendRows[Transpose[Jap],Jpp,Jpw],
                   AppendRows[Transpose[Jaw],Transpose[Jpw],Jww]];
          t = Table[Sqrt[Inverse[fisher][[k]][[k]]], {k,1,3*m}];
          res = Append[res,t];), {i,1,ntot}];
          Return[If[Head[Nsam] === List, res, res[[1]]]]]

LocalPeaks[p_List, n_Integer] :=
   Module[{i,pd,p2,val},
          p2 = Take[p,Length[p]/2];
          pd = Drop[p2,1] - Drop[p2,-1];
          i = Select[Range[Length[pd]-1],
              ((pd[[#]] >= 0.0)&&(pd[[#+1]] <= 0.0))&]+1;
          {val,i} = Transpose[Take[Reverse[Sort[Transpose[{p2[[i]],i}]]],n]];
          Return[{N[(Pi/Length[p2])*i],val}]]

LocalPeaks[data_List, zpfactor_Integer, window_, n_Integer]:=
   Module[{p},
          p = ZeroPad[data,zpfactor*Length[data]];
          p = Periodogram[p*window[Length[p]]];
          Return[LocalPeaks[p,n]]]

Options[SinMl] = {RelAcc->10.0^(-6),MaxIter->200,Method->Exact}
SinMl[data_List, init_List, opts___] :=
   Module[{ahat,cost,f,itmax,m,maxiter,method,Nsam,phihat,relacc,sighat,y,ymin},
          relacc = RelAcc /. {opts} /. Options[SinMl];
          itmax = MaxIter /. {opts} /. Options[SinMl];
          method = Method /. {opts} /. Options[SinMl];
          m = Length[init];
          Nsam = Length[data];
          f[y_] := 
             Module[{Hc,Hs,k,l,t,tempc,temps},
                    tempc = Table[Sum[data[[t+1]] Cos[y[[k]] (t - (Nsam-1)/2)],
                                  {t,0,Nsam-1}],{k,1,m}];
                    temps = Table[Sum[data[[t+1]] Sin[y[[k]] (t - (Nsam-1)/2)],
                                  {t,0,Nsam-1}],{k,1,m}];
                    If[method === Approx,
                       Return[-(tempc . tempc + temps . temps)]];
                    Hc = Table[Dirichlet[0.5*(y[[k]]-y[[l]]),Nsam],
                               {k,1,m},{l,1,m}];
                    Hs = Table[Dirichlet[0.5*(y[[k]]+y[[l]]),Nsam],
                               {k,1,m},{l,1,m}];
                    {Hc,Hs} = {2*Inverse[Hc+Hs],2*Inverse[Hc-Hs]};
                    Return[-(tempc . Hc . tempc + temps . Hs . temps)]];
          {ymin,costmin,iter} = Minimize2[f,init,relacc,itmax];
          {ahat,phihat,sighat} = EstRest[data,ymin];
          Return[{ahat,phihat,ymin,sighat}]]

EstRest[data_List, omega_List] :=
   Module[{ac,as,Hc,Hs,k,l,Nsam,sighat,t,tempc,temps},
          m = Length[omega];
          Nsam = Length[data];
          tempc = Table[Sum[data[[t+1]] Cos[omega[[k]] (t - (Nsam-1)/2)],
                        {t,0,Nsam-1}],{k,1,m}];
          temps = Table[Sum[data[[t+1]] Sin[omega[[k]] (t - (Nsam-1)/2)],
                        {t,0,Nsam-1}],{k,1,m}];
          Hc = Table[Dirichlet[0.5*(omega[[k]]-omega[[l]]),Nsam],
                        {k,1,m},{l,1,m}];
          Hs = Table[Dirichlet[0.5*(omega[[k]]+omega[[l]]),Nsam],
                        {k,1,m},{l,1,m}];
          {Hc,Hs} = {2*Inverse[Hc+Hs],2*Inverse[Hc-Hs]};
          {ac,as} = {Hc . tempc, Hs . temps};
          sighat = Sqrt[Nsam^(-1) (data . data - (tempc . ac + temps . as))];
          Return[{Sqrt[ac^2+as^2],ArcTan[as/ac],sighat}]]

Options[TruncatedSvd] = {Method->FCov}
TruncatedSvd[data_List, l_Integer, m_Integer, opts___] :=
   Module[{ahat,ind,mat,method,Nsam,r,s,sigu2,u,v,vec},
          method = Method /. {opts} /. Options[TruncatedSvd];
          Nsam = Length[data];
          {mat,vec} = Which[method === FCov,
                            {Toeplitz[Take[data,{l,Nsam-1}],
                                      Reverse[Take[data,l]]],
                             Take[data,{l+1,Nsam}]},
                            method === BCov,
                            {Hankel[Take[data,{2,Nsam-l+1}],
                                    Take[data,{Nsam-l+1,Nsam}]],
                             Take[data,{1,Nsam-l}]},
                            method === FBCov,
                            {Join[Toeplitz[Take[data,{l,Nsam-1}],
                                            Reverse[Take[data,l]]],
                                   Hankel[Take[data,{2,Nsam-l+1}],
                                          Take[data,{Nsam-l+1,Nsam}]]],
                              Join[Take[data,{l+1,Nsam}],
                                   Take[data,{1,Nsam-l}]]},
                            True, (Print["Unrecognized Method"]; {{},{}})];
          If[vec == {}, Return[Null]];
          {u,s,v} = SingularValues[mat];
          u = Take[u,2*m]; s = Take[s,2*m]; v = Take[v,2*m];
          ahat = Prepend[-Transpose[v] . (DiagonalMatrix[s^(-1)] . (u . vec)),
                         1];
          r = Proots[ahat];
          ind = Transpose[Sort[
                Transpose[{Abs[Abs[r]-1.0], Range[l]}]]][[2]];
          r = Arg[r[[Take[ind,2*m]]]];
          r = Select[r,(# > 0)&];
          {ahat,phihat,sighat} = EstRest[data,r];
          Return[{ahat,phihat,r,sighat}]]

ModYuleWalker[data_List, n_Integer, l_Integer, m_Integer] :=
   Module[{ahat,col,ind,phihat,r,R,rhat,row,sighat,x},
          rhat = CovSeq[data,n];
          row = Reverse[Take[rhat,{2,l+1}]];
          col = Take[rhat,{l+1,n}];
          R = Toeplitz[col,row];
          r = Take[rhat,{l+2,n+1}];
          x = If[Length[row] == Length[col],
                 Inverse[R] . r, LstSqr[R,r]];
          ahat = Prepend[-x,1];
          r = Proots[ahat];
          ind = Transpose[Sort[
                Transpose[{Abs[Abs[r]-1.0], Range[l]}]]][[2]];
          r = Arg[r[[Take[ind,2*m]]]];
          r = Select[r,(# > 0)&];
          {ahat,phihat,sighat} = EstRest[data,r];
          Return[{ahat,phihat,r,sighat}]]

Music[data_List, l_Integer, m_Integer] :=
   Module[{i,poly,r,R,s,u,v},
          r = CovSeq[data,l];
          R = Toeplitz[r,r];
          {u,s,v} = SingularValues[R];
          u = Drop[u,2*m];
          poly = Sum[Convolution[u[[i]],Reverse[u[[i]]]],{i,1,Length[u]}];
          r = Proots[poly];
          r = Select[r, (Abs[#] <= 1.0)&];
          ind = Transpose[Sort[
                Transpose[{Abs[Abs[r]-1.0], Range[l]}]]][[2]];
          r = Arg[r[[Take[ind,2*m]]]];
          r = Select[r,(# > 0)&];
          {ahat,phihat,sighat} = EstRest[data,r];
          Return[{ahat,r,sighat}]]

SetAttributes[Dirichlet, Listable]
Dirichlet[0, n_] := n
Dirichlet[0.0, n_] := n
Dirichlet[x_, n_] := Sin[n x]/Sin[x]

End[]
EndPackage[]
Null
